import warnings

from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor, ALPHABET, is_orix_rotation
from Elasticipy.tensors.stress_strain import StrainTensor
import numpy as np
from scipy.spatial.transform import Rotation

class ThermalExpansionTensor(SymmetricSecondOrderTensor):
    name = 'Thermal expansion tensor'

    def __mul__(self, other):
        if isinstance(other, Rotation) or is_orix_rotation(other):
            return super().__mul__(other)
        else:
            other = np.asarray(other)
            other_expanded = other[..., None, None]
            other_with_eye = other_expanded * np.ones(3)
            new_mat = self.matrix * other_with_eye
            return StrainTensor(new_mat)

    def apply_temperature(self, temperature, mode='pair'):
        """
        Apply temperature increase to the thermal expansion tensor, or to the array.

        Application can be made pair-wise, or considering all cross-combinations (see below).

        Parameters
        ----------
        temperature : float or numpy.ndarray
        mode : str, optional
            If "pair" (default), the temperatures are applied pair-wise on the tensor array. Broadcasting rule applies
            If "cross", all cross combinations are considered. Therefore, if ``C=A.apply_temperature(T, mode="cross")``,
            then ``C.shape=A.shape + T.shape``.

        Returns
        -------
        StrainTensor
            Strain corresponding to the applied temperature increase(s).

        Examples
        --------
        Let consider a transverse isotropic case (e.g. carbon fibers):
        >>> from Elasticipy.ThermalExpansion import ThermalExpansionTensor as ThEx
        >>> alpha = ThEx.transverse_isotropic(alpha_11=5.6e-6, alpha_33=-0.4e-6)

        Now apply temperature increases:
        >>> T = [0, 1, 2]
        >>> eps = alpha.apply_temperature(T)

        We get a strain tensor of the same shape as the applied temperatures:
        >>> eps
        Strain tensor
        Shape=(3,)
        >>> eps[-1]
        Strain tensor
        [[ 1.12e-05  0.00e+00  0.00e+00]
         [ 0.00e+00  1.12e-05  0.00e+00]
         [ 0.00e+00  0.00e+00 -8.00e-07]]

         Now let's rotate the thermal expansions:
         >>> from scipy.spatial.transform import Rotation
         >>> rot = Rotation.random(3) # Set of 3 random 3D rotations
         >>> alpha_rotated = alpha * rot

         If we want to combine each rotated thermal expansion with the corresponding temperature increase:
         >>> eps_rotated_pair = alpha_rotated * T # Equivalent to alpha_rotated.apply_temperature(T)
         >>> eps_rotated_pair
         Strain tensor
         Shape=(3,)

         Conversely, if we want to evaluate all cross-combinations of thermal expansions and temperature increases:

         >>> eps_rotated_cross = alpha_rotated.apply_temperature(T, mode='cross')
         >>> eps_rotated_cross
         Strain tensor
         Shape=(3, 3)

        """
        temperature = np.asarray(temperature)
        if mode == 'pair':
            matrix = self.matrix*temperature[...,np.newaxis,np.newaxis]
        elif mode == 'cross':
            indices_self = ALPHABET[:self.ndim]
            indices_temp = ALPHABET[:len(temperature.shape)].upper()
            ein_str = indices_self + 'ij,' + indices_temp + '->' + indices_self + indices_temp + 'ij'
            matrix = np.einsum(ein_str, self.matrix, temperature)
        else:
            raise ValueError('Invalid mode. It could be either "pair" or "cross".')
        return StrainTensor(matrix)


    def matmul(self, other):
        """
        Matrix like product with array of float, resulting either in StrainTensor.

        Compute the product between the tensor and a numpy array in a "matrix-product" way,*
        that is by computing each of the products. If T.shape=(m,n,o,...) and other.shape=(p,q,r,...), then::

            T.matmul(other).shape = (m,n,o,...,p,q,r,...)

        Parameters
        ----------
        other : np.ndarray
            Value to multiply by.
        Returns
        -------
        StrainTensor
            Array fo strain tensors corresponding to all cross-combinations between Thermal expansions and temperature
            increases.
        """
        warnings.warn(
            'matmul() is deprecated and will be removed in a future version. Use dot(tensor,mode="cross") or '
            'rotate(rotation,mode="cross") instead.',
            DeprecationWarning,
            stacklevel=2)
        if isinstance(other, Rotation) or is_orix_rotation(other):
            return super().matmul(other)
        else:
            return self.apply_temperature(other, mode='cross')


    @classmethod
    def isotropic(cls, alpha):
        """
        Create an isotropic thermal expansion tensor.

        Parameters
        ----------
        alpha : float
            Thermal expansion coefficient.

        Returns
        -------
        ThermalExpansionTensor
        """
        return cls(np.eye(3)*alpha)

    @classmethod
    def orthotropic(cls, alpha_11, alpha_22, alpha_33):
        """
        Create an orthotropic thermal expansion tensor.

        Parameters
        ----------
        alpha_11, alpha_22, alpha_33 : float
            Thermal expansion coefficients along the first, second and third axes, respectively.

        Returns
        -------
        ThermalExpansionTensor
        """
        return cls(np.diag([alpha_11, alpha_22, alpha_33]))

    @classmethod
    def orthorhombic(cls, *args):
        """
        Create a thermal expansion tensor corresponding to an orthotropic thermal expansion coefficient.

        This function is an alias for orthotropic().

        Parameters
        ----------
        args : list
            Orthotropic thermal expansion coefficient.

        Returns
        -------
        ThermalExpansionTensor

        See Also
        --------
        orthotropic
        """
        return cls.orthotropic(*args)

    @classmethod
    def monoclinic(cls, alpha_11, alpha_22, alpha_33, alpha_13=None, alpha_12=None):
        """
        Create a thermal expansion tensor for monoclinic symmetry.

        If alpha_13, the Diad || z is assumed. If alpha_12, the Diad || z is assumed. Therefore, these two parameters
        are exclusive.

        Parameters
        ----------
        alpha_11, alpha_22, alpha_33 : float
            Thermal expansion coefficient along the first, second and third axes, respectively.
        alpha_13 : float, optional
            Thermal expansion coefficient corresponding to XZ shear (for Diad || y)
        alpha_12: float, optional
            Thermal expansion coefficient corresponding to XY shear (for Diad || z)

        Returns
        -------
        ThermalExpansionTensor
        """
        matrix = np.diag([alpha_11, alpha_22, alpha_33])
        if (alpha_13 is not None) and (alpha_12 is None):
            matrix[0, 2] = matrix[2, 0]= alpha_13
        elif (alpha_12 is not None) and (alpha_13 is None):
            matrix[0, 1] = matrix[1, 0]= alpha_12
        elif (alpha_13 is not None) and (alpha_12 is not None):
            raise ValueError('alpha_13 and alpha_12 cannot be used together.')
        else:
            raise ValueError('Either alpha_13 or alpha_12 must be provided.')
        return cls(matrix)


    @classmethod
    def triclinic(cls, alpha_11=0., alpha_12=0., alpha_13=0., alpha_22=0., alpha_23=0., alpha_33=0.):
        """
        Create a thermal expansion tensor for triclinic symmetry.

        Parameters
        ----------
        alpha_11, alpha_12, alpha_13, alpha_22, alpha_23, alpha_33 : float
            Values of the thermal expansion coefficients

        Returns
        -------
        ThermalExpansionTensor
        """
        mat = [[alpha_11, alpha_12, alpha_13],
               [alpha_12, alpha_22, alpha_23],
               [alpha_13, alpha_23, alpha_33]]
        return cls(mat)

    @classmethod
    def transverse_isotropic(cls, alpha_11, alpha_33):
        """
        Create a thermal expansion tensor for transverse isotropic symmetry.

        Parameters
        ----------
        alpha_11 : float
            Thermal expansion coefficient along the first and second axes
        alpha_33 : float
            Thermal expansion coefficient along the third axis

        Returns
        -------
        ThermalExpansionTensor
        """
        return cls(np.diag([alpha_11, alpha_11, alpha_33]))

    @property
    def volumetric_coefficient(self):
        """ Returns the volumetric thermal expansion coefficient."""
        return self.I1


from Elasticipy.tensors.elasticity import StiffnessTensor

C = StiffnessTensor.from_MP("mp-1048")
C.save_to_txt("TiNi.txt")
C.save_to_txt("TiNi-matrix.txt", matrix_only=True)

C2 = StiffnessTensor.from_txt_file("TiNi.txt")
C3 = StiffnessTensor.from_txt_file("TiNi-matrix.txt")
print(C==C2)
print(C==C3)





import unittest
from Elasticipy.interfaces.FEPX import from_step_file, from_results_folder
from Elasticipy.tensors.second_order import SecondOrderTensor, SymmetricSecondOrderTensor, \
    SkewSymmetricSecondOrderTensor
from Elasticipy.tensors.stress_strain import StrainTensor, StressTensor
import os
import numpy as np

current_dir = os.path.dirname(os.path.abspath(__file__))
FEPX_DATA = os.path.join(current_dir,'interfaces/FEPX/simulation.sim/results/elts/')
SIZE_FEPX_DATA = 2453
NSTEP_FEPX_DATA = 3

class TestFEPX(unittest.TestCase):
    def test_stress_from_file(self):
        a = from_step_file(FEPX_DATA + 'strain/strain.step0')
        assert isinstance(a, StrainTensor)
        assert a.shape == (SIZE_FEPX_DATA,)

    def test_strain_from_file(self):
        a = from_step_file(FEPX_DATA + 'stress/stress.step0')
        assert isinstance(a, StressTensor)
        assert a.shape == (SIZE_FEPX_DATA,)

    def test_stress_from_folder(self):
        a = from_results_folder(FEPX_DATA + 'strain')
        assert isinstance(a, StrainTensor)
        assert a.shape == (NSTEP_FEPX_DATA, SIZE_FEPX_DATA)

    def test_defrate_from_folder(self):
        a = from_results_folder(FEPX_DATA + 'defrate')
        assert isinstance(a, SymmetricSecondOrderTensor)
        assert a.shape == (NSTEP_FEPX_DATA, SIZE_FEPX_DATA)

    def test_velgrad_from_folder(self):
        a = from_results_folder(FEPX_DATA + 'velgrad')
        assert isinstance(a, SecondOrderTensor) and not isinstance(a, SymmetricSecondOrderTensor)
        assert a.shape == (NSTEP_FEPX_DATA, SIZE_FEPX_DATA)

    def test_spinrate_from_folder(self):
        a = from_results_folder(FEPX_DATA + 'spinrate')
        assert isinstance(a, SkewSymmetricSecondOrderTensor)
        assert a.shape == (NSTEP_FEPX_DATA, SIZE_FEPX_DATA)

    def test_orientation_from_folder(self):
        a = from_results_folder(FEPX_DATA + 'ori')
        assert isinstance(a, np.ndarray)
        assert a.shape == (NSTEP_FEPX_DATA, SIZE_FEPX_DATA, 3)

if __name__ == '__main__':
    unittest.main()



import unittest
from Elasticipy.interfaces.PRISMS import from_quadrature_file, from_stressstrain_file
from Elasticipy.tensors.second_order import SecondOrderTensor, SymmetricSecondOrderTensor
from Elasticipy.tensors.stress_strain import StressTensor
import numpy as np
import os
import pandas as pd

current_dir = os.path.dirname(os.path.abspath(__file__))
quadrature_file = os.path.join(current_dir, 'interfaces/PRISMS/QuadratureOutputs.csv')
stressstrain_file = os.path.join(current_dir, 'interfaces/PRISMS/stressstrain.txt')

quadrature_data = pd.read_csv(quadrature_file, header=None, usecols=range(0,37))
stressstrain_data = pd.read_csv(stressstrain_file, sep='\t')

class TestPRISMSInterfaces(unittest.TestCase):
    def test_from_quadrature(self):
        stress = from_quadrature_file(quadrature_file)
        assert stress.shape == (len(quadrature_data),)
        assert isinstance(stress, StressTensor)
        for i in range(0,len(quadrature_data)):
            assert stress[i].C[0, 0] == quadrature_data.iloc[i,28]
            assert stress[i].C[1, 1] == quadrature_data.iloc[i,29]
            assert stress[i].C[2, 2] == quadrature_data.iloc[i,30]
            assert stress[i].C[0, 1] == quadrature_data.iloc[i,31]
            assert stress[i].C[0, 2] == quadrature_data.iloc[i,32]
            assert stress[i].C[1, 2] == quadrature_data.iloc[i,34]

    def test_from_stressstrain_with_fields(self):
        fields = ('grain ID', 'phase ID', 'det(J)', 'twin', 'coordinates', 'orientation', 'elastic gradient',
                  'plastic gradient', 'stress')
        a = from_quadrature_file(quadrature_file, returns=fields)
        assert len(a) == len(fields)
        assert isinstance(a[0], np.ndarray)
        assert isinstance(a[1], np.ndarray)
        assert isinstance(a[2], np.ndarray)
        assert isinstance(a[3], np.ndarray)
        assert isinstance(a[4], np.ndarray)
        assert isinstance(a[5], np.ndarray)
        assert isinstance(a[6], SecondOrderTensor)
        assert isinstance(a[7], SecondOrderTensor)
        assert isinstance(a[8], StressTensor)

    def test_from_stressstrain(self):
        E, stress = from_stressstrain_file(stressstrain_file)
        assert E.shape == stress.shape == (len(stressstrain_data),)
        assert isinstance(E, SymmetricSecondOrderTensor)
        assert isinstance(stress, StressTensor)
        for i in range(0,len(stressstrain_data)):
            assert E[i].C[0, 0] == stressstrain_data['Exx'][i]
            assert E[i].C[1, 1] == stressstrain_data['Eyy'][i]
            assert E[i].C[2, 2] == stressstrain_data['Ezz'][i]
            assert E[i].C[1, 2] == stressstrain_data['Eyz'][i]
            assert E[i].C[0, 2] == stressstrain_data['Exz'][i]
            assert E[i].C[0, 1] == stressstrain_data['Exy'][i]
            assert stress[i].C[0, 0] == stressstrain_data['Txx'][i]
            assert stress[i].C[1, 1] == stressstrain_data['Tyy'][i]
            assert stress[i].C[2, 2] == stressstrain_data['Tzz'][i]
            assert stress[i].C[1, 2] == stressstrain_data['Tyz'][i]
            assert stress[i].C[0, 2] == stressstrain_data['Txz'][i]
            assert stress[i].C[0, 1] == stressstrain_data['Txy'][i]



if __name__ == '__main__':
    unittest.main()



import numpy as np
from Elasticipy.tensors.stress_strain import StrainTensor, StressTensor


class IsotropicHardening:
    """
    Template class for isotropic hardening plasticity models
    """
    type = "Isotropic"
    name = 'Generic'

    def __init__(self, criterion='von Mises'):
        """
        Create an instance of a plastic model, assuming isotropic hardening

        Parameters
        ----------
        criterion : str or PlasticityCriterion
            Plasticity criterion to use. Can be 'von Mises', 'Tresca' or 'J2'. J2 is the same as von Mises.
        """
        if isinstance(criterion, str):
            criterion = criterion.lower()
            if criterion in ('von mises', 'mises', 'vonmises', 'j2'):
                self.criterion = VonMisesPlasticity
            elif criterion == 'tresca':
                self.criterion = TrescaPlasticity
            else:
                raise ValueError('The criterion can be "Tresca", "von Mises" or "J2".')
        else:
            self.criterion = criterion
        self.plastic_strain = 0.0

    def __repr__(self):
        return (('{} plasticity model\n'.format(self.name) +
                ' type: {}\n'.format(self.type)) +
                ' criterion: {}\n'.format(self.criterion.name) +
                ' current strain: {}'.format(self.plastic_strain))

    def flow_stress(self, strain, **kwargs):
        pass

    def apply_strain(self, strain, **kwargs):
        """
        Apply strain to the current JC model.

        This function updates the internal variable to store hardening state.

        Parameters
        ----------
        strain : float or StrainTensor
        kwargs : dict
            Keyword arguments passed to flow_stress()

        Returns
        -------
        float
            Associated flow stress (positive)

        See Also
        --------
        flow_stress : compute the flow stress, given a cumulative equivalent strain
        """
        if isinstance(strain, float):
            self.plastic_strain += np.abs(strain)
        elif isinstance(strain, StrainTensor):
            self.plastic_strain += strain.eq_strain()
        else:
            raise ValueError('The applied strain must be float of StrainTensor')
        return self.flow_stress(self.plastic_strain, **kwargs)

    def compute_strain_increment(self, stress, **kwargs):
        pass

    def reset_strain(self):
        self.plastic_strain = 0.0


class JohnsonCook(IsotropicHardening):
    """
    Special case of isotropic hardening with an underlying Johnson Cook hardening evolution rule
    """
    name = "Johnson-Cook"

    def __init__(self, A, B, n, C=None, eps_dot_ref=1.0, m=None, T0=25, Tm=None, criterion='von Mises'):
        """
        Constructor for a Johnson-Cook (JC) model.

        The JC model is an exponential-law strain hardening model, which can take into account strain-rate sensibility
        and temperature-dependence (although they are not mandatory). See notes for details.

        Parameters
        ----------
        A : float
            Yield stress
        B : float
            Work hardening coefficient
        n : float
            Work hardening exponent
        C : float, optional
            Strain-rate sensitivity coefficient
        eps_dot_ref : float, optional
            Reference strain-rate
        m : float, optional
            Temperature sensitivity exponent
        T0 : float, optional
            Reference temperature
        Tm : float, optional
            Melting temperature (at which the flow stress is zero)
        criterion : str or PlasticityCriterion, optional
            Plasticity criterion to use. It can be 'von Mises' or 'Tresca'.

        Notes
        -----
        The flow stress (:math:`\\sigma`) depends on the strain (:math:`\\varepsilon`),
        the strain rate :math:`\\dot{\\varepsilon}` and
        the temperature (:math:`T`) so that:

        .. math::

                \\sigma = \\left(A + B\\varepsilon^n\\right)
                        \\left(1 + C\\log\\left(\\frac{\\varepsilon}{\\dot{\\varepsilon}_0}\\right)\\right)
                        \\left(1-\\theta^m\\right)

        with

        .. math::

                \\theta = \\begin{cases}
                            \\frac{T-T_0}{T_m-T_0} & \\text{if } T<T_m\\\\
                            1                      & \\text{otherwise}
                            \\end{cases}
        """
        super().__init__(criterion=criterion)
        self.A = A
        self.B = B
        self.C = C
        self.n = n
        self.m = m
        self.eps_dot_ref = eps_dot_ref
        self.T0 = T0
        self.Tm = Tm

    def flow_stress(self, eps_p, eps_dot=None, T=None):
        """
        Compute the flow stress from the Johnson-Cook model

        Parameters
        ----------
        eps_p : float or list or tuple or numpy.ndarray
            Equivalent plastic strain
        eps_dot : float or list or tuple or numpy.ndarray, optional
            Equivalent plastic strain rate. If float, the strain-rate is supposed to be homogeneous for every value of
            eps_p.
        T : float or list or tuple or np.ndarray
            Temperature. If float, the temperature is supposed to be homogeneous for every value of eps_p.
        Returns
        -------
        float or numpy.ndarray
            Flow stress
        """
        eps_p = np.asarray(eps_p)
        stress = (self.A + self.B * eps_p**self.n)

        if eps_dot is not None:
            eps_dot = np.asarray(eps_dot)
            if (self.C is None) or (self.eps_dot_ref is None):
                raise ValueError('C and eps_dot_ref must be defined for using a rate-dependent model')
            stress *= (1 + self.C * np.log(eps_dot / self.eps_dot_ref))

        if T is not None:
            T = np.asarray(T)
            if self.T0 is None or self.Tm is None or self.m is None:
                raise ValueError('T0, Tm and m must be defined for using a temperature-dependent model')
            theta = (T - self.T0) / (self.Tm - self.T0)
            theta = np.clip(theta, None, 1.0)
            stress *= (1 - theta**self.m)

        return stress



    def compute_strain_increment(self, stress, T=None, apply_strain=True, criterion='von Mises'):
        """
        Given the equivalent stress, compute the strain increment with respect to the normality rule.

        Parameters
        ----------
        stress : float or StressTensor
            Equivalent stress to compute the stress from, or full stress tensor.
        T : float
            Temperature
        apply_strain : bool, optional
            If true, the JC model will be updated to account for the applied strain (hardening)
        criterion : str, optional
            Plasticity criterion to consider to compute the equivalent stress and apply the normality rule.
            It can be 'von Mises', 'Tresca' or 'J2'. 'J2' is equivalent to 'von Mises'.

        Returns
        -------
        StrainTensor or float
            Increment of plastic strain. If the input stress is float, only the magnitude of the increment will be
            returned (float value). If the stress is of type StressTensor, the returned value will be a full
            StrainTensor.

        See Also
        --------
        apply_strain : apply strain to the JC model and updates its hardening value
        """
        if isinstance(stress, StressTensor):
            eq_stress = self.criterion.eq_stress(stress)
        else:
            eq_stress = stress
        if T is None:
            if eq_stress > self.A:
                k = eq_stress  - self.A
                total_strain = (1 / self.B * k) ** (1 / self.n)
                strain_increment = np.max((total_strain - self.plastic_strain, 0))
            else:
                strain_increment = 0.0
        else:
            if self.T0 is None or self.Tm is None or self.m is None:
                raise ValueError('T0, Tm and m must be defined for using a temperature-dependent model')
            else:
                if T >= self.Tm:
                    strain_increment = np.inf
                else:
                    theta = (T - self.T0) / (self.Tm - self.T0)
                    theta_m = theta**self.m
                    k = (eq_stress / (1 - theta_m) - self.A)
                    if k<0:
                        strain_increment = 0.0
                    else:
                        total_strain = (1/self.B * k)**(1/self.n)
                        strain_increment = np.max((total_strain - self.plastic_strain, 0))
        if apply_strain:
            self.apply_strain(strain_increment)

        if isinstance(stress, StressTensor):
            n = self.criterion.normal(stress)
            return n * strain_increment
        else:
            return strain_increment

    def reset_strain(self):
        """
        Reinitialize the plastic strain to 0
        """
        self.plastic_strain = 0.0


class PlasticityCriterion:
    """
    Template class for plasticity criteria
    """
    name = 'generic'

    @staticmethod
    def eq_stress(stress, **kwargs):
        """
        Return the equivalent stress, with respect to the plasticity criterion.

        Parameters
        ----------
        stress : StressTensor
            Stress to compute the equivalent stress from
        kwargs : dict
            keyword arguments passed to the function
        Returns
        -------
        float or numpy.ndarray
        """
        pass

    def normal(self, stress, **kwargs):
        """
        Apply the normality rule

        Parameters
        ----------
        stress : StressTensor
            Stress tensor to apply the normality rule
        kwargs : dict
            Keyword arguments passed to the function

        Returns
        -------
        StrainTensor
            Normalized direction of plastic flow
        """
        pass

class VonMisesPlasticity(PlasticityCriterion):
    """
    von Mises plasticity criterion, with associated normality rule
    """
    name = 'von Mises'
    @staticmethod
    def eq_stress(stress, **kwargs):
        return stress.vonMises()

    @staticmethod
    def normal(stress, **kwargs):
        eq_stress = stress.vonMises()
        dev_stress = stress.deviatoric_part()
        gradient_tensor = dev_stress / eq_stress
        return StrainTensor(3 / 2 * gradient_tensor.matrix)

class TrescaPlasticity(PlasticityCriterion):
    """
    Tresca plasticity criterion, with associated normality rule
    """
    name = 'Tresca'

    @staticmethod
    def eq_stress(stress, **kwargs):
        return stress.Tresca()

    @staticmethod
    def normal(stress, **kwargs):
        vals, dirs = stress.eig()
        u1 = dirs[..., 0]
        u3 = dirs[..., 2]
        s1 = vals[..., 0]
        s2 = vals[..., 1]
        s3 = vals[..., 2]
        a = np.einsum('...i,...j->...ij', u1, u1)
        b = np.einsum('...i,...j->...ij', u3, u3)
        normal = a - b
        singular_points = np.logical_or(s2 == s1, s2 == s3)
        normal[singular_points] = VonMisesPlasticity().normal(stress[singular_points]).matrix
        normal[np.logical_and(s2 == s1, s2 == s3)] = 0.0
        strain = StrainTensor(normal)
        return strain / strain.eq_strain()

class DruckerPrager(PlasticityCriterion):
    """
    Drucker-Prager pressure-dependent plasticity criterion, with associated normality rule
    """
    name = 'Drucker-Prager'

    def __init__(self, alpha):
        """
        Create a Drucker-Prager (DG) plasticity criterion.

        Parameters
        ----------
        alpha : float
            Pressure dependence parameters (see notes for details)

        Notes
        -----
        The pressure-dependent DG plasticity criterion assumes that the equivalent stress is defined as:


        """
        self.alpha = alpha

    def eq_stress(self, stress, **kwargs):
        return (stress.J2**0.5 + self.alpha * stress.I1) / (1/3**0.5 + self.alpha)

    def normal(self, stress, **kwargs):
        J2 = stress.J2
        gradient = stress.deviatoric_part() / (2 * J2**0.5) + self.alpha * StressTensor.eye(stress.shape)
        strain = StrainTensor(gradient.matrix)
        return strain / strain.eq_strain()


import numpy as np

a = np.sqrt(2)
KELVIN_MAPPING_MATRIX = np.array([[1, 1, 1, a, a, a],
                                  [1, 1, 1, a, a, a],
                                  [1, 1, 1, a, a, a],
                                  [a, a, a, 2, 2, 2],
                                  [a, a, a, 2, 2, 2],
                                  [a, a, a, 2, 2, 2], ])

VOIGT_MAPPING_MATRIX_COMPLIANCE = [[1, 1, 1, 2, 2, 2],
                                  [1, 1, 1, 2, 2, 2],
                                  [1, 1, 1, 2, 2, 2],
                                  [2, 2, 2, 4, 4, 4],
                                  [2, 2, 2, 4, 4, 4],
                                  [2, 2, 2, 4, 4, 4]]

class MappingConvention:
    matrix = np.array(KELVIN_MAPPING_MATRIX)

    @property
    def mapping_inverse(self):
        return self

class KelvinMapping(MappingConvention):
    name = 'Kelvin'

class VoigtMapping(MappingConvention):
    name = 'Voigt'

    def __init__(self, tensor='Stiffness'):
        if tensor == 'Stiffness':
            self.matrix = np.ones((6,6))
            self.tensor_type = 'Stiffness'
        else:
            self.matrix = np.array(VOIGT_MAPPING_MATRIX_COMPLIANCE)
            self.tensor_type = 'Compliance'

    @property
    def mapping_inverse(self):
        if self.tensor_type == 'Stiffness':
            return VoigtMapping(tensor='Compliance')
        else:
            return VoigtMapping(tensor='Stiffness')


from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor, SecondOrderTensor, \
    SkewSymmetricSecondOrderTensor
from Elasticipy.tensors.stress_strain import StressTensor, StrainTensor
import pandas as pd
import numpy as np
import os
import re
from pathlib import Path

DTYPES={'stress':StressTensor,
        'strain':StrainTensor,
        'strain_el':StrainTensor,
        'strain_pl':StrainTensor,
        'velgrad':SecondOrderTensor,
        'defrate':SymmetricSecondOrderTensor,
        'defrate_pl':SymmetricSecondOrderTensor,
        'spinrate':SkewSymmetricSecondOrderTensor,
        }


def _list_valid_filenames(folder, startswith='strain'):
    file_list = os.listdir(folder)
    pattern = r'{}\.step\d+'.format(startswith)
    return [f for f in file_list if re.fullmatch(pattern, f)]

def from_step_file(file, dtype=None):
    """
    Import data from a single step file given by FEPX.

    The type of returns is inferred from the data one wants to parse
    (according the `FEPX documentation <https://fepx.info/doc/output.html>`_ ).

    Parameters
    ----------
    file : str
        Path to the file to read
    dtype : type, optional
        If provided, sets the type of returned array. It can be:
          - float
          - SecondOrderTensor
          - SymmetricSecondOrderTensor
          - SkewSymmetricSecondOrderTensor
          - stressTensor
          - strainTensor

    Returns
    -------
    SecondOrderTensor or numpy.ndarray
        Array of second-order tensors built from the read data. The array will be of shape (n,), where n is the number
        of elements in the mesh.
    """
    data = pd.read_csv(file, header=None, sep=' ')
    array = data.to_numpy()
    base_name = os.path.splitext(os.path.basename(file))[0]
    if dtype is None:
        if base_name in DTYPES:
            dtype = DTYPES[base_name]
        else:
            dtype = float
    if issubclass(dtype,SymmetricSecondOrderTensor):
        return dtype.from_Voigt(array, voigt_map=[1,1,1,1,1,1])
    elif dtype == SkewSymmetricSecondOrderTensor:
        zeros = np.zeros(array.shape[0])
        mat = np.array([[ zeros,         array[:, 0],   array[:, 1]],
                        [-array[:, 0],  zeros,          array[:, 2]],
                        [-array[:, 1], -array[:, 2],    zeros     ]]).transpose((2, 0, 1))
        return SkewSymmetricSecondOrderTensor(mat)
    elif dtype == SecondOrderTensor:
        length = array.shape[0]
        return SecondOrderTensor(array.reshape((length,3,3)))
    elif dtype == float:
        return array



def from_results_folder(folder, dtype=None):
    """
    Import all data of a given field from FEPX results folder.

    The type of returns is inferred from the data one wants to parse
    (according the `FEPX documentation <https://fepx.info/doc/output.html>`_ ).

    Parameters
    ----------
    folder : str
        Path to the folder to read the results from
    dtype : type, optional
        If provided, sets the type of returned array. It can be:
          - float
          - SecondOrderTensor
          - SymmetricSecondOrderTensor
          - SkewSymmetricSecondOrderTensor
          - stressTensor
          - strainTensor

    Returns
    -------
    SecondOrderTensor or numpy.ndarray
        Array of second-order tensors built from the read data. The array will be of shape (m, n), where m is the number
         of time increment n is the number of elements in the mesh.
    """
    dir_path = Path(folder)
    folder_name = dir_path.name
    if not dir_path.is_dir():
        raise ValueError(f"{folder} is not a valid directory.")
    constructor = None
    array = []
    for file in dir_path.iterdir():
        if file.is_file() and file.name.startswith(folder_name):
            data_file = from_step_file(str(file), dtype=dtype)
            if constructor is None:
                constructor = type(data_file)
            elif constructor != type(data_file):
                raise ValueError('The types of data contained in {} seem to be inconsistent.'.format(folder))
            array.append(data_file)
    if constructor == np.ndarray:
        return np.stack(array, axis=0)
    else:
        return constructor.stack(array)



import numpy as np
from matplotlib import pyplot as plt, cm
from matplotlib.colors import Normalize
from numpy import cos, sin
from scipy import integrate as integrate
from scipy import optimize
from Elasticipy.polefigure import add_polefigure


def sph2cart(*args):
    """
    Converts spherical/hyperspherical coordinates to cartesian coordinates.

    Parameters
    ----------
    args : tuple
        (phi, theta) angles for spherical coordinates of direction u, where phi denotes the azimuth from X and theta is
        the colatitude angle from Z.
        If a third argument is passed, it defines the third angle in hyperspherical coordinate system (psi), that is
        the orientation of the second vector v, orthogonal to u.

    Returns
    -------
    numpy.ndarray
        directions u expressed in cartesian coordinates system.
    tuple of numpy.ndarray, numpy.ndarray
        If a third angle is passed, returns a tuple:
        - The first element is `u`, the directions expressed in the cartesian coordinate system.
        - The second element is `v`, the direction of the second vector orthogonal to `u`, also expressed in the
        cartesian coordinate system.
    """
    phi, theta, *psi = args
    phi_vec = np.array(phi).flatten()
    theta_vec = np.array(theta).flatten()
    u = np.array([cos(phi_vec) * sin(theta_vec), sin(phi_vec) * sin(theta_vec), cos(theta_vec)]).T
    if not psi:
        return u
    else:
        psi_vec = np.array(psi).flatten()
        e_phi = np.array([-sin(phi_vec), cos(phi_vec), np.zeros(phi_vec.shape)])
        e_theta = np.array([cos(theta_vec) * cos(phi_vec), cos(theta_vec) * sin(phi_vec), -sin(theta_vec)])
        v = cos(psi_vec) * e_phi + sin(psi_vec) * e_theta
        return u, v.T


def _plot3D(fig, u, r, **kwargs):
    norm = Normalize(vmin=r.min(), vmax=r.max())
    colors = cm.viridis(norm(r))
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    xyz = (u.T * r.T).T
    ax.plot_surface(xyz[:, :, 0], xyz[:, :, 1], xyz[:, :, 2], facecolors=colors, rstride=1, cstride=1,
                    antialiased=False, **kwargs)
    mappable = cm.ScalarMappable(cmap='viridis', norm=norm)
    mappable.set_array([])
    fig.colorbar(mappable, ax=ax)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    return ax


def _create_xyz_section(ax, section_name, polar_angle):
    ax.title.set_text('{}-{} plane'.format(*section_name))
    if section_name == 'XY':
        phi = polar_angle
        theta = np.pi / 2 * np.ones(len(polar_angle))
    elif section_name == 'XZ':
        phi = np.zeros(len(polar_angle))
        theta = np.pi / 2 - polar_angle
    else:
        phi = (np.pi / 2) * np.ones(len(polar_angle))
        theta = np.pi / 2 - polar_angle
    ax.set_xticks(np.linspace(0, 3 * np.pi / 2, 4))
    h_direction, v_direction = section_name
    ax.set_xticklabels((h_direction, v_direction, '-' + h_direction, '-' + v_direction))
    return phi, theta, ax


def uniform_spherical_distribution(n_evals, seed=None, return_orthogonal=False):
    """
    Create a set of vectors whose projections over the unit sphere are uniformly distributed.

    Parameters
    ----------
    n_evals : int
        Number of vectors to generate
    seed : int, default None
        Sets the seed for the random values. Useful if one wants to ensure reproducibility.
    return_orthogonal : bool, default False
        If true, also return a second set of vectors which are orthogonal to the first one.

    Returns
    -------
    u : np.ndarray
        Random set of vectors whose projections over the unit sphere are uniform.
    v : np.ndarray
        Set of vectors with the same properties as u, but orthogonal to u.
        Returned only if return_orthogonal is True

    Notes
    -----
    The returned vector(s) are not unit. If needed, one can use:
        u = (u.T / np.linalg.norm(u, axis=1)).T
    """
    if seed is None:
        rng = np.random
    else:
        rng = np.random.default_rng(seed)
    u = rng.normal(size=(n_evals, 3))
    u /= np.linalg.norm(u, axis=1, keepdims=True)
    if return_orthogonal:
        if seed is not None:
            # Ensure that the seed used for generated v is not the same as that for u
            # Otherwise, u and v would be equal.
            rng = np.random.default_rng(seed+1)
        u2 = rng.normal(size=(n_evals, 3))
        return u, np.cross(u, u2)
    else:
        return u

def _integrate_over_unit_sphere(phi, theta, values=None, psi=None):
    sine = np.sin(theta)
    if values is None:
        values = np.ones(phi.shape)
    if psi is None:
        return integrate.trapezoid(
                    integrate.trapezoid(
                        values * sine, axis=0, x=phi[:, 0]
                    ), axis=0, x=theta[0, :])
    else:
        return integrate.trapezoid(
                    integrate.trapezoid(
                        integrate.trapezoid(
                            values * sine, axis=0, x=phi[:, 0, 0]
                        ), axis=0, x=theta[0, :, 0])
                , x=psi[0,0,:])



class SphericalFunction:
    """
    Class for spherical functions, that is, functions that depend on directions in 3D space.

    Attribute
    ---------
    fun : function to use
    """
    domain = np.array([[0, 2 * np.pi],
                       [0, np.pi / 2]])
    name = 'Spherical function'
    """Bounds to consider in spherical coordinates"""

    def __init__(self, fun, symmetry=True):
        """
        Create a spherical function, that is, a function that depends on one direction only.

        Parameters
        ----------
        fun : callable
            Function to return
        symmetry : bool, optional
            Set to true if fun(u)==fun(-u)
        """
        self.fun = fun
        self.symmetry = symmetry

    def __repr__(self):
        val_min, _ = self.min()
        val_max, _ = self.max()
        s = '{}\n'.format(self.name)
        s += 'Min={}, Max={}'.format(val_min, val_max)
        return s

    def __add__(self, other):
        if type(other) is self.__class__:
            def fun(*x):
                return self.fun(*x) + other.fun(*x)
            return self.__class__(fun)
        elif isinstance(other, (float, int, np.number)):
            def fun(*x):
                return self.fun(*x) + other
            return self.__class__(fun)
        else:
            msg_error = 'A {} can only be added to another {} or a scalar value.'.format(self.name, self.name)
            raise NotImplementedError(msg_error)

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            def fun(*x):
                return self.fun(*x) - other.fun(*x)
            return self.__class__(fun)
        else:
            return self.__add__(-other)

    def __mul__(self, other):
        if type(other) is self.__class__:
            def fun(*x):
                return self.fun(*x) * other.fun(*x)
        elif isinstance(other, (float, int, np.number)):
            def fun(*x):
                return self.fun(*x) * other
        else:
            msg_error = 'A {} can only be multiplied by another {} or a scalar value.'.format(self.name, self.name)
            raise NotImplementedError(msg_error)
        return self.__class__(fun)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if type(other) is self.__class__:
            def fun(*x):
                return self.fun(*x) / other.fun(*x)
        elif isinstance(other, (float, int, np.number)):
            def fun(*x):
                return self.fun(*x) / other
        else:
            raise NotImplementedError('A SphericalFunction can only be divided by a scalar value of another SphericalFunction.')
        return self.__class__(fun)

    def eval(self, u):
        """
        Evaluate value along a given (set of) direction(s).

        Parameters
        ----------
        u : np.ndarray or list
            Direction(s) to estimate the value along with. It can be of a unique direction [nx, ny, nz],
            or a set of directions (e.g. [[n1x, n1y n1z],[n2x, n2y, n2z],...]).

        Returns
        -------
        float or np.ndarray
            If only one direction is given as a tuple of floats [nx, ny, nz], the result is a float; otherwise, the
            result is a nd.array.

        See Also
        --------
        eval_spherical : evaluate the function along a given direction given using the spherical coordinates
        """
        u_vec = np.atleast_2d(u)
        norm = np.linalg.norm(u_vec, axis=1)
        if np.any(norm < 1e-9):
            raise ValueError('The input vector cannot be zeros')
        u_vec = (u_vec.T / norm).T
        values = self.fun(u_vec)
        if isinstance(u, list) and np.array(u).shape == (3,):
            return values[0]
        else:
            return values

    def eval_spherical(self, *args, degrees=False):
        """
        Evaluate value along a given (set of) direction(s) defined by its (their) spherical coordinates.

        Parameters
        ----------
        args : list or np.ndarray
            [phi, theta] where phi denotes the azimuth angle from X axis,
            and theta is the latitude angle from Z axis (theta==0 -> Z axis).
        degrees : bool, default False
            If True, the angles are given in degrees instead of radians.

        Returns
        -------
        float or np.ndarray
            If only one direction is given as a tuple of floats [phi, theta], the result is a float; otherwise, the
            result is a nd.array.

        See Also
        --------
        eval : evaluate the function along a direction given by its cartesian coordinates
        """
        angles = np.atleast_2d(args)
        if degrees:
            angles = np.radians(angles)
        phi, theta = angles.T
        u = sph2cart(phi, theta)
        values = self.eval(u)
        if (np.array(args).shape == (2,) or np.array(args).shape == (1, 2)) and not isinstance(args, np.ndarray):
            return values[0]
        else:
            return values

    def _global_minimizer(self, fun):
        n_eval = 50
        phi = np.linspace(*self.domain[0], n_eval)
        theta = np.linspace(*self.domain[1], n_eval)
        if len(self.domain) == 2:
            phi, theta = np.meshgrid(phi, theta)
            angles0 = np.array([phi.flatten(), theta.flatten()]).T
        else:
            psi = np.linspace(*self.domain[2], n_eval)
            phi, theta, psi = np.meshgrid(phi, theta, psi)
            angles0 = np.array([phi.flatten(), theta.flatten(), psi.flatten()]).T
        values = fun(angles0)
        loc_x0 = np.argmin(values)
        angles1 = angles0[loc_x0]
        results = optimize.minimize(fun, angles1, method='L-BFGS-B', bounds=self.domain)
        return results

    def min(self):
        """
        Find minimum value of the function.

        Returns
        -------
        fmin : float
            Minimum value
        dir : np.ndarray
            Direction along which the minimum value is reached

        See Also
        --------
        max : return the maximum value and the location where it is reached
        """
        results = self._global_minimizer(self.eval_spherical)
        return results.fun, sph2cart(*results.x)

    def max(self):
        """
        Find maximum value of the function.

        Returns
        -------
        min : float
            Maximum value
        direction : np.ndarray
            direction along which the maximum value is reached

        See Also
        --------
        min : return the minimum value and the location where it is reached
        """
        def fun(x):
            return -self.eval_spherical(*x)

        results = self._global_minimizer(fun)
        return -results.fun, sph2cart(*results.x)

    def mean(self, method='trapezoid', n_evals=10000, seed=None):
        """
        Estimate the mean value along all directions in the 3D space.

        Parameters
        ----------
        method : str {'exact', 'Monte Carlo', 'trapezoid'}, optional
            If 'exact', the full integration is performed over the unit sphere (see Notes). If 'trapezoid', the trapeze
            method is used to approximate the integral. If 'Monte Carlo', the function is evaluated along a finite set
            of random directions.
        n_evals : int, optional
            If method=='Monte Carlo' or 'trapezoid', sets the number of unit directions to use.
        seed : int, default None
            Sets the seed for random sampling when using the Monte Carlo method. Useful when one wants to reproduce
            results.

        Returns
        -------
        float
            Mean value

        See Also
        --------
        std : Standard deviation of the spherical function over the unit sphere.

        Notes
        -----
        The full integration over the unit sphere, used if method=='exact', takes advantage of numpy.integrate.dblquad.
        This algorithm is robust but usually slow. The Monte Carlo method can be 1000 times faster.
        """
        if self.symmetry:
            dom_size = 2 * np.pi
        else:
            dom_size = 4 * np.pi
        if method == 'exact':
            def fun(theta, phi):
                return self.eval_spherical(phi, theta) * sin(theta)

            domain = self.domain.flatten()
            q = integrate.dblquad(fun, *domain)
            return q[0] / dom_size
        elif method == 'trapezoid':
            (phi, theta), evals = self.evaluate_on_spherical_grid(n_evals)
            dom_size = _integrate_over_unit_sphere(phi, theta)
            return _integrate_over_unit_sphere(phi, theta, values=evals) / dom_size
        else:
            u = uniform_spherical_distribution(n_evals, seed=seed)
            return np.mean(self.eval(u))

    def var(self, method='trapezoid', n_evals=10000, mean=None, seed=None):
        """
        Estimate the variance along all directions in the 3D space

        Parameters
        ----------
        method : str {'exact', 'Monte Carlo', 'trapezoid}, optional
            If 'exact', the full integration is performed over the unit sphere (see Notes). If 'trapezoid', the trapeze
            method is used to approximate the integral. If 'Monte Carlo', the function is evaluated along a finite set
            of random directions.
        n_evals : int, optional
            If method=='Monte Carlo' or 'trapezoid', sets the number of unit directions to use.
        mean : float, optional
            If provided, skip estimation of mean value and use that provided instead (only used for exact and trapezoid
            methods)
        seed : int, optional
            Sets the seed for random sampling when using the Monte Carlo method. Useful when one wants to reproduce
            results.

        Returns
        -------
        float
            Variance of the function

        See Also
        --------
        mean : mean value of the function over the unit sphere.

        Notes
        -----
        The full integration over the unit sphere, used if method=='exact', takes advantage of numpy.integrate.dblquad.
        This algorithm is robust but usually slow. The Monte Carlo method can be 1000 times faster.
        """
        if method == 'exact':
            if mean is None:
                mean = self.mean(method='exact')

            def fun(theta, phi):
                return (self.eval_spherical(phi, theta) - mean) ** 2 * sin(theta)

            domain = self.domain.flatten()
            q = integrate.dblquad(fun, *domain)
            return q[0] / (2 * np.pi)
        elif method == 'trapezoid':
            if mean is None:
                mean = self.mean(method="trapezoid", n_evals=n_evals)
            (phi, theta), evals = self.evaluate_on_spherical_grid(n_evals)
            dom_size = _integrate_over_unit_sphere(phi, theta)
            return _integrate_over_unit_sphere(phi, theta, values=(evals - mean)**2) / dom_size
        else:
            u = uniform_spherical_distribution(n_evals, seed=seed)
            return np.var(self.eval(u))

    def std(self, **kwargs):
        """
        Standard deviation of the function along all directions in the 3D space.

        Parameters
        ----------
        **kwargs
            These parameters will be passed to var() function

        Returns
        -------
        float
            Standard deviation

        See Also
        --------
        var : variance of the function
        """
        return np.sqrt(self.var(**kwargs))

    def evaluate_on_spherical_grid(self, n, return_in_spherical=True, use_symmetry=True):
        """
        Create a set of vectors corresponding to a spherical grid (phi,theta), then flatten it.

        Parameters
        ----------
        n : int or tuple of int
            If int, it give the overall number of evaluations over the quarter unit sphere. If tuple of int, they
            correspond to the number of spherical angles (n_phi, n_theta).
        return_in_spherical : bool, optional
            If true, the first output argument will be the spherical coordinates (phi, theta). Otherwise, the cartesian
            coordinates are returned
        use_symmetry : whether to take consider the upper half-domain only, or the full sphere.

        Returns
        -------
        numpy.ndarray or tuple
            Coordinates of evaluation. If return_in_spherical==True, they will be returned as a tuple of angle (phi,
            theta). Otherwise, they will be returned as a numpy array of shape (n_phi, n_theta, 3).
        numpy.ndarray
            Grid of evaluated values of shape (n_phi, n_theta, 3).
        """
        symmetry = self.symmetry and use_symmetry
        if isinstance(n, int):
            if symmetry:
                n_theta = int(np.sqrt(n)/ 2) + 1
                n_phi = 4 * n_theta
            else:
                n_theta = int(np.sqrt(n / 2)) + 1
                n_phi = 2 * n_theta
        else:
            n_phi, n_theta = n
        if symmetry:
            theta_max = np.pi / 2
        else:
            theta_max = np.pi
        phi = np.linspace(0, 2 * np.pi, n_phi)
        theta = np.linspace(0, theta_max, n_theta)
        phi_grid, theta_grid = np.meshgrid(phi, theta, indexing='ij')
        u = sph2cart(phi_grid.flatten(), theta_grid.flatten())
        evals = self.eval(u)
        evals_grid = evals.reshape((n_phi, n_theta))
        if return_in_spherical:
            return (phi_grid, theta_grid), evals_grid
        else:
            return u.reshape((n_phi, n_theta, 3)), evals_grid

    def plot3D(self, n_phi=50, n_theta=50, fig=None, **kwargs):
        """
        3D plotting of a spherical function

        Parameters
        ----------
        n_phi : int, default 50
            Number of azimuth angles (phi) to use for plotting. Default is 50.
        n_theta : int, default 50
            Number of latitude angles (theta) to use for plotting. Default is 50.
        fig : matplotlib.figure.Figure, default None
            handle to existing figure object. If None, a new figure will be created. If passed, the figure is not shown,
            (one should use plt.show() afterward).
        **kwargs
            These parameters will be passed to matplotlib plot_surface() function.

        Returns
        -------
        matplotlib.figure.Figure
            Handle to the figure
        matplotlib.Axes3D
            Handle to axes

        See Also
        --------
        plot_xyz_sections : plot values of the function in X-Y, X-Z an Y-Z planes.
        """
        if fig is None:
            new_fig = plt.figure()
        else:
            new_fig = fig
        u, evals = self.evaluate_on_spherical_grid((n_phi, n_theta), return_in_spherical=False, use_symmetry=False)
        ax = _plot3D(new_fig, u, evals, **kwargs)
        ax.axis('equal')
        return new_fig, ax

    def plot_xyz_sections(self, n_theta=500, fig=None, axs=None, **kwargs):
        """
        Plot XYZ sections of spherical data.

        This method generates a figure with three polar plots showing the XY, XZ, and YZ sections of the
        spherical data contained within the instance. Each section is plotted using n_theta evenly spaced
        points over the interval [0, 2π].

        Parameters
        ----------
        n_theta : int, optional
            Number of points to use for the polar plot angles. Default is 500.
        fig : matplotlib.figure.Figure, default None
            Handle to existing figure object. If None, a new figure will be created. If passed, the figure is not shown
        axs : tuple of matplotlib.projections.polar.PolarAxes, optional
            If provided, use these axes to plot the sections, instead of creating new ones.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the plot function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object containing the polar plots.
        axs : list of matplotlib.axes._subplots.PolarAxesSubplot
            List of axes objects for each plot.
        """
        if fig is None:
            new_fig = plt.figure()
        else:
            new_fig = fig
        theta_polar = np.linspace(0, 2*np.pi, n_theta)
        titles = ('XY', 'XZ', 'YZ')
        axs_new = []
        for i in range(0, 3):
            if axs is None:
                ax = new_fig.add_subplot(1, 3, i+1, projection='polar')
            else:
                ax = axs[i]
            angles = np.zeros((n_theta, 2))
            phi, theta, ax = _create_xyz_section(ax, titles[i], theta_polar)
            angles[:, 0] = phi
            angles[:, 1] = theta
            r = self.eval_spherical(angles)
            ax.plot(theta_polar, r, **kwargs)
            axs_new.append(ax)
        return new_fig, axs_new

    def plot_as_pole_figure(self, n_theta=50, n_phi=200, projection='lambert',
                            fig=None, plot_type='imshow', title=None,
                            subplot_args=(), subplot_kwargs=None, **kwargs):
        """
        Plots a pole figure visualization of spherical data using specified parameters and plot types.

        The function creates a pole figure based on spherical coordinates and can use different
        types of plots including 'imshow', 'contourf', and 'contour'. It also supports a variety of
        projections and renders the figure using Matplotlib.

        Parameters
        ----------
        n_theta : int, optional
            Number of divisions in the theta dimension, by default 50.
        n_phi : int, optional
            Number of divisions in the phi dimension, by default 200.
        projection : str {'Lambert','equal area'}, optional
            The type of projection to use for the plot, by default 'Lambert'.
        fig : matplotlib.figure.Figure, optional
            A Matplotlib figure object. If None, a new figure will be created, by default None.
        plot_type : str, optional
            The type of plot to generate: 'imshow', 'contourf', or 'contour', by default 'imshow'.
        title : str, optional
            Title to add to the current axis. Default is None.
        subplot_args : tuple
            List of arguments to pass to the subplot function, by default ()
        subplot_kwargs : dict
            Dictionary of keyword-arguments to pass to the subplot function, by default {}
        **kwargs : dict
            Additional keyword arguments to pass to the plotting functions.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The Matplotlib figure containing the plot.
        ax : matplotlib.axes._axes.Axes
            The Matplotlib axes of the plot.
        """
        if subplot_kwargs is None:
            subplot_kwargs = {}
        if fig is None:
            new_fig = plt.figure()
        else:
            new_fig = fig
        ax = add_polefigure(new_fig, *subplot_args, projection=projection, **subplot_kwargs)
        phi = np.linspace(*self.domain[0], n_phi)
        theta = np.linspace(*self.domain[1], n_theta)
        phi, theta = np.meshgrid(phi, theta)
        phi_flat = phi.flatten()
        theta_flat = theta.flatten()
        values = self.eval_spherical(np.array([phi_flat, theta_flat]).T)
        if plot_type == 'imshow':
            sc = ax.pcolormesh(phi, theta, values.reshape(phi.shape), **kwargs)
        elif plot_type == 'contourf':
            sc = ax.contourf(phi, theta, values.reshape(phi.shape), **kwargs)
        elif plot_type == 'contour':
            sc = ax.contour(phi, theta, values.reshape(phi.shape), **kwargs)
        else:
            raise ValueError(f'Unknown plot type: {plot_type}')
        ax.set_rlim(*self.domain[1])
        ax.set_title(title)
        new_fig.colorbar(sc)
        return new_fig, ax


class HyperSphericalFunction(SphericalFunction):
    """
    Class for defining functions that depend on two orthogonal directions u and v.
    """
    domain = np.array([[0, 2 * np.pi],
                       [0, np.pi / 2],
                       [0, np.pi]])
    name = 'Hyperspherical function'

    def __init__(self, fun):
        """
        Create a hyperspherical function, that is, a function that depends on two orthogonal directions only.
        """
        super().__init__(fun)

    def eval(self, u, *args):
        """
        Evaluate the Hyperspherical function with respect to two orthogonal directions.

        Parameters
        ----------
        u : list or np.ndarray
            First axis
        args : list or np.ndarray
            Second axis

        Returns
        -------
        float or np.ndarray
            Function value

        See Also
        --------
        eval_spherical : evaluate the function along a direction defined by its spherical coordinates.
        """
        m_vec = np.atleast_2d(u)
        n_vec = np.atleast_2d(*args)
        norm_1 = np.linalg.norm(m_vec, axis=1)
        norm_2 = np.linalg.norm(n_vec, axis=1)
        if np.any(norm_1 < 1e-9) or np.any(norm_2 < 1e-9):
            raise ValueError('The input vector cannot be zeros')
        m_vec = (m_vec.T / norm_1).T
        n_vec = (n_vec.T / norm_2).T
        dot = np.abs(np.einsum('ij,ij->i', m_vec, n_vec))
        if np.any(dot > 1e-9):
            raise ValueError('The two directions must be orthogonal.')
        values = self.fun(m_vec, n_vec)
        if np.array(u).shape == (3,) and not isinstance(u, np.ndarray):
            return values[0]
        else:
            return values

    def mean(self, method='trapezoid', n_evals=int(1e6), seed=None):
        if method == 'exact':
            def fun(psi, theta, phi):
                return self.eval_spherical(phi, theta, psi) * sin(theta)

            domain = self.domain.flatten()
            q = integrate.tplquad(fun, *domain)
            return q[0] / (2 * np.pi ** 2)
        elif method == 'trapezoid':
            (phi, theta, psi), evals = self.evaluate_on_spherical_grid(n_evals)
            dom_size = _integrate_over_unit_sphere(phi, theta, psi=psi)
            integral = _integrate_over_unit_sphere(phi, theta, psi=psi, values=evals)
            return integral / dom_size
        else:
            u, v = uniform_spherical_distribution(n_evals, seed=seed, return_orthogonal=True)
            return np.mean(self.eval(u, v))

    def eval_spherical(self, *args, degrees=False):
        """
        Evaluate value along a given (set of) direction(s) defined by its (their) spherical coordinates.

        Parameters
        ----------
        args : list or np.ndarray
            [phi, theta, psi] where phi denotes the azimuth angle from X axis to the first direction (u), theta is
            the latitude angle from Z axis (theta==0 -> u = Z axis), and psi is the angle defining the orientation of
            the second direction (v) in the plane orthogonal to u, as illustrated below:

            .. image:: ../../docs/_static/images/HyperSphericalCoordinates.png


        degrees : bool, default False
            If True, the angles are given in degrees instead of radians.

        Returns
        -------
        float or np.ndarray
            If only one direction is given as a tuple of floats [nx, ny, nz], the result is a float;
        otherwise, the result is a nd.array.

        See Also
        --------
        eval : evaluate the function along two orthogonal directions (u,v))
        """
        angles = np.atleast_2d(args)
        if degrees:
            angles = np.radians(angles)
        phi, theta, psi = angles.T
        u, v = sph2cart(phi, theta, psi)
        values = self.eval(u, v)
        if np.array(args).shape == (3,) and not isinstance(args, np.ndarray):
            return values[0]
        else:
            return values

    def var(self, method='trapezoid', n_evals=int(1e6), mean=None, seed=None):
        if method == 'exact':
            if mean is None:
                mean = self.mean(method='exact')

            def fun(psi, theta, phi):
                return (mean - self.eval_spherical(phi, theta, psi)) ** 2 * sin(theta)

            domain = self.domain.flatten()
            q = integrate.tplquad(fun, *domain)
            return q[0] / (2 * np.pi ** 2)
        if method == 'trapezoid':
            (phi, theta, psi), evals = self.evaluate_on_spherical_grid(n_evals)
            dom_size = _integrate_over_unit_sphere(phi, theta, psi=psi)
            if mean is None:
                mean = self.mean(method='trapezoid', n_evals=n_evals)
            return _integrate_over_unit_sphere(phi, theta, psi=psi, values=(evals - mean)**2) / dom_size
        else:
            u, v = uniform_spherical_distribution(n_evals, seed=seed, return_orthogonal=True)
            return np.var(self.eval(u, v))

    def evaluate_on_spherical_grid(self, n, return_in_spherical=True, use_symmetry=True):
        """
        Create a set of vectors corresponding to a spherical grid (phi,theta), then flatten it.

        Parameters
        ----------
        n : int or tuple of int
            If int, gives the overall number of evaluations over the unit hypersphere. If a tuple is passed, they gieve
            the number of angles to consider for (hyper)spherical coordinates (n_phi, n_theta, n_psi).
        return_in_spherical : bool, optional
            If true, the first output argument will be the spherical coordinates (phi, theta). Otherwise, the cartersian
            coordinates are returned
        use_symmetry : bool, optional
            Whether to use take advantage ot symmetry

        Returns
        -------
        tuple
            Coordinates of evaluation, either in spherical of cartesian coordinates
        numpy.ndarray
            Grid of evaluated values
        """
        symmetry = self.symmetry and use_symmetry
        if isinstance(n, int):
            if symmetry:
                n_phi = int(2 * n ** (1 / 3)) + 1
                n_theta = int(n_phi / 4) + 1
                n_psi = int(n_phi / 2) + 1
            else:
                n_phi = int(4 * n ** (1 / 3)) + 1
                n_theta = int(n_phi / 2) + 1
                n_psi = int(n_phi / 2) + 1
        else:
            n_phi, n_theta, n_psi = n
        if symmetry:
            theta_max = np.pi / 2
        else:
            theta_max = np.pi
        phi = np.linspace(0, 2 * np.pi, n_phi)
        theta = np.linspace(0, theta_max, n_theta)
        psi = np.linspace(0, np.pi, n_psi)
        phi_grid, theta_grid, psi_grid = np.meshgrid(phi, theta, psi, indexing='ij')
        u, v = sph2cart(phi_grid.flatten(), theta_grid.flatten(), psi_grid)
        evals = self.eval(u, v)
        evals_grid = evals.reshape((n_phi, n_theta, n_psi))
        if return_in_spherical:
            return (phi_grid, theta_grid, psi_grid), evals_grid
        else:
            u_r = u.reshape((n_phi, n_theta, n_psi, 3))
            v_r = v.reshape((n_phi, n_theta, n_psi, 3))
            return (u_r, v_r), evals_grid

    def plot3D(self, n_phi=50, n_theta=50, n_psi=50, which='mean', fig=None, **kwargs):
        """
        Generate a 3D plot representing the evaluation of spherical harmonics.

        This function evaluates a function over a grid defined by spherical coordinates
        (phi, theta, psi) and produces a 3D plot. It provides options to display the mean,
        standard deviation, minimum, or maximum of the evaluated values along the third angles (psi).
        The plot can be customized with additional keyword arguments.

        Parameters
        ----------
        n_phi : int, optional
            Number of divisions along the phi axis, default is 50.
        n_theta : int, optional
            Number of divisions along the theta axis, default is 50.
        n_psi : int, optional
            Number of divisions along the psi axis, default is 50.
        which : str, optional
            Determines which statistical measure to plot ('mean', 'std', 'min', 'max'),
            default is 'mean'.
        fig : matplotlib.figure.Figure, optional
            Handle to existing figure object. Default is None. If provided, it disables showing the figure.
        kwargs : dict, optional
            Additional keyword arguments to customize the plot.

        Returns
        -------
        tuple
            A tuple containing the matplotlib figure and axes objects.
        """
        if fig is None:
            new_fig = plt.figure()
        else:
            new_fig = fig
        uv, values = self.evaluate_on_spherical_grid((n_phi, n_theta, n_psi), return_in_spherical=False, use_symmetry=False)
        u, _ = uv
        if which == 'std':
            r_grid = np.std(values, axis=2)
        elif which == 'min':
            r_grid = np.min(values, axis=2)
        elif which == 'max':
            r_grid = np.max(values, axis=2)
        else:
            r_grid = np.mean(values, axis=2)
        ax = _plot3D(new_fig, u[:, :, 0, :], r_grid, **kwargs)
        return new_fig, ax

    def plot_xyz_sections(self, n_theta=500, n_psi=100, color_minmax='blue', alpha_minmax=0.2, color_mean='red',
                          fig=None):
        """
        Plots the XYZ sections using polar projections.

        This function creates a figure with three subplots representing the XY, XZ,
        and YZ sections. It utilizes polar projections to plot the min, max, and mean
        values of the evaluated function over given theta and phi ranges.

        Parameters
        ----------
        n_theta : int, optional
            Number of theta points to use in the grid (default is 500).
        n_psi : int, optional
            Number of psi points to use in the grid (default is 100).
        color_minmax : str, optional
            Color to use for plotting min and max values (default is 'blue').
        alpha_minmax : float, optional
            Alpha transparency level to use for the min/max fill (default is 0.2).
        color_mean : str, optional
            Color to use for plotting mean values (default is 'red').
        fig : matplotlib.figure.Figure, optional
            Handle to existing figure object. Default is None. If provided, it disables showing the figure.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The created figure.
        axs : list of matplotlib.axes._subplots.PolarAxesSubplot
            List of polar axis subplots.
        """
        if fig is None:
            new_fig = plt.figure()
        else:
            new_fig = fig
        theta_polar = np.linspace(0, 2 * np.pi, n_theta)
        titles = ('XY', 'XZ', 'YZ')
        handles, labels = [], []
        axs = []
        for i in range(0, 3):
            ax = new_fig.add_subplot(1, 3, i+1, projection='polar')
            phi, theta, ax = _create_xyz_section(ax, titles[i], theta_polar)
            psi = np.linspace(0, np.pi, n_psi)
            phi_grid, psi_grid = np.meshgrid(phi, psi, indexing='ij')
            theta_grid, _ = np.meshgrid(theta, psi, indexing='ij')
            phi = phi_grid.flatten()
            theta = theta_grid.flatten()
            psi = psi_grid.flatten()
            u, v = sph2cart(phi, theta, psi)
            values = self.eval(u, v).reshape((n_theta, n_psi))
            min_val = np.min(values, axis=1)
            max_val = np.max(values, axis=1)
            ax.plot(theta_polar, min_val, color=color_minmax)
            ax.plot(theta_polar, max_val, color=color_minmax)
            area = ax.fill_between(theta_polar, min_val, max_val, alpha=alpha_minmax, label='Min/Max')
            line, = ax.plot(theta_polar, np.mean(values, axis=1), color=color_mean, label='Mean')
            axs.append(ax)
        handles.extend([line, area])
        labels.extend([line.get_label(), area.get_label()])
        new_fig.legend(handles, labels, loc='upper center', ncol=2, bbox_to_anchor=(0.5, 0.95))
        return new_fig, axs

    def plot_as_pole_figure(self, n_theta=50, n_phi=200, n_psi=50, which='mean', projection='lambert', fig=None,
                            plot_type='imshow', show=True, title=None, subplot_args=(), subplot_kwargs=None, **kwargs):
        """
        Generate a pole figure plot from spherical function evaluation.

        This function evaluates a spherical function over specified ranges of angles
        (phi, theta, psi) and then generates a 2D pole figure plot using various
        statistical summaries of the data (mean, std, min, max). It also supports
        several types of plot visualizations such as 'imshow', 'contourf', and 'contour'.

        Parameters
        ----------
        n_theta : int, optional
            Number of sampling points for theta angle. Default is 50.
        n_phi : int, optional
            Number of sampling points for phi angle. Default is 200.
        n_psi : int, optional
            Number of sampling points for psi angle. Default is 50.
        which : str, optional
            Specifies the type of statistical summary to use for plotting.
            Options include 'mean', 'std', 'min', 'max'. Default is 'mean'.
        projection : str, optional
            Type of projection for the pole figure plot. Default is 'lambert'.
        fig : matplotlib.figure.Figure, optional
            Pre-existing figure to use for plotting. If None, a new figure is created.
            Default is None.
        plot_type : str, optional
            Type of plot to generate. Can be 'imshow', 'contourf', or 'contour'.
            Default is 'imshow'.
        show : bool, optional
            Set whether to show the plot or not. Default is True. This must be turned off when using multiple subplots.
        subplot_args : tuple, optional
            Arguments to pass to the add_subplot() function. Default is None.
        subplot_kwargs : dict, optional
            Keyword arguments to pass to the add_subplot() function. Default is None.
        **kwargs : dict, optional
            Additional keyword arguments passed to the plotting functions.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object containing the plot.
        ax : matplotlib.axes.Axes
            The axes object containing the plot.
        """
        if subplot_kwargs is None:
            subplot_kwargs = {}
        if fig is None:
            fig = plt.figure()
        ax = add_polefigure(fig, *subplot_kwargs, projection=projection, **subplot_kwargs)
        phi = np.linspace(*self.domain[0], n_phi)
        theta = np.linspace(*self.domain[1], n_theta)
        psi = np.linspace(*self.domain[2], n_psi)
        phi_grid, theta_grid, psi_grid = np.meshgrid(phi, theta, psi, indexing='ij')
        phi_flat = phi_grid.flatten()
        theta_flat = theta_grid.flatten()
        psi_flat = psi_grid.flatten()
        values = self.eval_spherical(np.array([phi_flat, theta_flat, psi_flat]).T)
        reshaped_values = values.reshape((n_phi, n_theta, n_psi))
        if which == 'std':
            to_plot = np.std(reshaped_values, axis=2)
        elif which == 'min':
            to_plot = np.min(reshaped_values, axis=2)
        elif which == 'max':
            to_plot = np.max(reshaped_values, axis=2)
        else:
            to_plot = np.mean(reshaped_values, axis=2)
        phi_grid, theta_grid = np.meshgrid(phi, theta, indexing='ij')
        if plot_type == 'imshow':
            sc = ax.pcolormesh(phi_grid, theta_grid, to_plot, **kwargs)
        elif plot_type == 'contourf':
            sc = ax.contourf(phi_grid, theta_grid, to_plot, **kwargs)
        elif plot_type == 'contour':
            sc = ax.contour(phi_grid, theta_grid, to_plot, **kwargs)
        else:
            raise ValueError(f'Unknown plot type: {plot_type}')
        ax.set_rlim(*self.domain[1])
        ax.set_title(title)
        fig.colorbar(sc)
        return fig, ax


from Elasticipy.tensors.fourth_order import SymmetricFourthOrderTensor
from Elasticipy.spherical_function import SphericalFunction, HyperSphericalFunction
from Elasticipy.crystal_symmetries import SYMMETRIES
from Elasticipy.tensors.stress_strain import StrainTensor, StressTensor
from Elasticipy.tensors.mapping import VoigtMapping, KelvinMapping
import numpy as np
import re
from warnings import warn

def _parse_tensor_components(prefix, **kwargs):
    pattern = r'^{}(\d{{2}})$'.format(prefix)
    value = dict()
    for k, v in kwargs.items():
        match = re.match(pattern, k)  # Extract 'C11' to '11' and so
        if match:
            value[match.group(1)] = v
    return value

def _indices2str(ij):
    return f'{ij[0] + 1}{ij[1] + 1}'

def _isotropic_matrix(C11, C12, C44):
    return np.array([[C11, C12, C12, 0, 0, 0],
                     [C12, C11, C12, 0, 0, 0],
                     [C12, C12, C11, 0, 0, 0],
                     [0, 0, 0, C44, 0, 0],
                     [0, 0, 0, 0, C44, 0],
                     [0, 0, 0, 0, 0, C44]])


def _check_definite_positive(mat):
    try:
        np.linalg.cholesky(mat)
    except np.linalg.LinAlgError:
        eigen_val = np.linalg.eigvals(mat)
        raise ValueError('The input matrix is not definite positive (eigenvalues: {})'.format(eigen_val))

def _switch_poisson_ratios(nu_xy, nu_yx, Ex, Ey, indices):
    if nu_yx is None and nu_xy is not None:
        return nu_xy * Ey / Ex
    elif nu_yx is not None and nu_xy is None:
        return nu_yx
    else:
        raise ValueError('Either nu_{0}{1} or nu_{1}{0} must be provided'.format(*indices))

class StiffnessTensor(SymmetricFourthOrderTensor):
    """
    Class for manipulating fourth-order stiffness tensors.
    """
    tensor_name = 'Stiffness'
    _C11_C12_factor = 0.5
    _C46_C56_factor = 1.0
    _component_prefix = 'C'

    def __init__(self, M, symmetry='Triclinic', check_positive_definite=True, phase_name= None, mapping=VoigtMapping(), **kwargs):
        """
        Construct of stiffness tensor from a (6,6) matrix.

        The input matrix must be symmetric, otherwise an error is thrown (except if check_symmetry==False, see below)

        Parameters
        ----------
        M : np.ndarray
            (6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation, or array of shape
            (3,3,3,3).
        phase_name : str, default None
            Name to display
        symmetry : str, default Triclinic
            Name of the crystal's symmetry
        check_symmetry : bool, optional
            Whether to check or not that the input matrix is symmetric.
        force_symmetry : bool, optional
            If true, the major symmetry of the tensor is forces
        mapping : str or MappingConvention
            mapping convention to use. Default is VoigtMapping.

        Notes
        -----
        The units used when building the stiffness tensor are up to the user (GPa, MPa, psi etc.). Therefor, the
        results you will get when performing operations (Young's modulus, "product" with strain tensor etc.) will be
        consistent with these units. For instance, if the stiffness tensor is defined in GPa, the computed stress will
        be given in GPa as well.
        """
        super().__init__(M, mapping=mapping, **kwargs)
        if check_positive_definite:
            _check_definite_positive(self._matrix)
        self.symmetry = symmetry
        self.phase_name = phase_name

    def __mul__(self, other):
        if isinstance(other, StrainTensor):
            new_tensor = self.ddot(other)
            return StressTensor(new_tensor.matrix)
        elif isinstance(other, StressTensor):
            raise ValueError('You cannot multiply a stiffness tensor with a Stress tensor.')
        else:
            return super().__mul__(other)

    def __repr__(self):
        string = super().__repr__()
        if self.phase_name is not None:
            string += '\nPhase: {}'.format(self.phase_name)
        string += '\nSymmetry: {}'.format(self.symmetry)
        return string

    def inv(self):
        """
        Compute the reciprocal compliance tensor

        Returns
        -------
        ComplianceTensor
            Reciprocal tensor
        """
        C = np.linalg.inv(self._matrix)
        return ComplianceTensor(C, symmetry=self.symmetry, phase_name=self.phase_name)

    @classmethod
    def from_txt_file(cls, filename):
        """
        Load the tensor from a text file.

        The two first lines can have data about phase name and symmetry, but this is not mandatory.

        Parameters
        ----------
        filename : str
            Filename to load the tensor from.

        Returns
        -------
        SymmetricFourthOrderTensor
            The reconstructed tensor read from the file.

        See Also
        --------
        save_to_txt : create a tensor from text file

        """
        with open(filename, 'r') as f:
            lines = f.readlines()

        # Initialize defaults
        phase_name = None
        symmetry = 'Triclinic'
        matrix_start_index = 0

        # Parse phase name if available
        if lines and lines[0].startswith("Phase Name:"):
            phase_name = lines[0].split(": ", 1)[1].strip()
            matrix_start_index += 1

        # Parse symmetry if available
        if len(lines) > matrix_start_index and lines[matrix_start_index].startswith("Symmetry:"):
            symmetry = lines[matrix_start_index].split(": ", 1)[1].strip()
            matrix_start_index += 1

        # Parse matrix
        matrix = np.loadtxt(lines[matrix_start_index:])

        # Return the reconstructed object
        return cls(matrix, phase_name=phase_name, symmetry=symmetry)

    def save_to_txt(self, filename, matrix_only=False):
        """
        Save the tensor to a text file.

        Parameters
        ----------
        filename : str
            Filename to save the tensor to.
        matrix_only : bool, False
            If true, only the components of tje stiffness tensor is saved (no data about phase nor symmetry)

        See Also
        --------
        from_txt_file : create a tensor from text file

        """
        with open(filename, 'w') as f:
            if not matrix_only:
                if self.phase_name is not None:
                    f.write(f"Phase Name: {self.phase_name}\n")
                f.write(f"Symmetry: {self.symmetry}\n")
            for row in self._matrix:
                f.write("  " + "  ".join(f"{value:8.2f}" for value in row) + "\n")

    @classmethod
    def _matrixFromCrystalSymmetry(cls, symmetry='Triclinic', point_group=None, diad='y', prefix=None, **kwargs):
        if prefix is None:
            prefix = cls._component_prefix
        values = _parse_tensor_components(prefix, **kwargs)
        C = np.zeros((6, 6))
        symmetry = symmetry.capitalize()
        if ((symmetry == 'tetragonal') or (symmetry == 'trigonal')) and (point_group is None):
            raise ValueError('For tetragonal and trigonal symmetries, the point group is mandatory.')
        tetra_1 = ['4', '-4', '4/m']
        tetra_2 = ['4mm', '-42m', '422', '4/mmm']
        trigo_1 = ['3', '-3']
        trigo_2 = ['32', '-3m', '3m']
        if point_group is not None:
            if (point_group in tetra_1) or (point_group in tetra_2):
                symmetry = 'Tetragonal'
            elif (point_group in trigo_1) or (point_group in trigo_2):
                symmetry = 'Trigonal'
        symmetry_description = SYMMETRIES[symmetry]
        if symmetry == 'Tetragonal':
            if point_group in tetra_1:
                symmetry_description = symmetry_description[', '.join(tetra_1)]
            else:
                symmetry_description = symmetry_description[', '.join(tetra_2)]
        elif symmetry == 'Trigonal':
            if point_group in trigo_1:
                symmetry_description = symmetry_description[', '.join(trigo_1)]
            else:
                symmetry_description = symmetry_description[', '.join(trigo_2)]
        elif symmetry == 'Monoclinic':
            symmetry_description = symmetry_description["Diad || " + diad]
        for required_field in symmetry_description.required:
            C[required_field] = values[_indices2str(required_field)]

        # Now apply relationships between components
        for equality in symmetry_description.equal:
            for index in equality[1]:
                C[index] = C[equality[0]]
        for opposite in symmetry_description.opposite:
            for index in opposite[1]:
                C[index] = -C[opposite[0]]
        C11_C12 = symmetry_description.C11_C12
        if C11_C12:
            for index in C11_C12:
                C[index] = (C[0, 0] - C[0, 1]) * cls._C11_C12_factor

        if symmetry == 'Trigonal':
            C[3, 5] = cls._C46_C56_factor * C[3, 5]
            C[4, 5] = cls._C46_C56_factor * C[4, 5]

        return C + np.tril(C.T, -1)

    @classmethod
    def fromCrystalSymmetry(cls, symmetry='Triclinic', point_group=None, diad='y', phase_name=None, prefix=None,
                             **kwargs):
        """
        Create a fourth-order tensor from limited number of components, taking advantage of crystallographic symmetries

        Parameters
        ----------
        symmetry : str, default Triclinic
            Name of the crystallographic symmetry
        point_group : str
            Point group of the considered crystal. Only used (and mandatory) for tetragonal and trigonal symmetries.
        diad : str {'x', 'y'}, default 'x'
            Alignment convention. Sets whether x||a or y||b. Only used for monoclinic symmetry.
        phase_name : str, default None
            Name to use when printing the tensor
        prefix : str, default None
            Define the prefix to use when providing the components. By default, it is 'C' for stiffness tensors, 'S' for
            compliance.
        kwargs
            Keywords describing all the necessary components, depending on the crystal's symmetry and the type of tensor.
            For Stiffness, they should be named as 'Cij' (e.g. C11=..., C12=...).
            For Comliance, they should be named as 'Sij' (e.g. S11=..., S12=...).
            See examples below. The behaviour can be overriten with the prefix option (see above)

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        isotropic : creates an isotropic stiffness tensor from two paremeters (e.g. E and v).

        Notes
        -----
        The relationships between the tensor's components depend on the crystallogrpahic symmetry [1]_.

        References
        ----------
        .. [1] Nye, J. F. Physical Properties of Crystals. London: Oxford University Press, 1959.

        Examples
        --------
        >>> from Elasticipy.tensors.elasticity import StiffnessTensor\n
        >>> StiffnessTensor.fromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
        ...                                     C11=231, C12=127, C13=104,
        ...                                     C22=240, C23=131, C33=175,
        ...                                     C44=81, C55=11, C66=85,
        ...                                     C15=-18, C25=1, C35=-3, C46=3)
        Stiffness tensor (in Voigt mapping):
        [[231. 127. 104.   0. -18.   0.]
         [127. 240. 131.   0.   1.   0.]
         [104. 131. 175.   0.  -3.   0.]
         [  0.   0.   0.  81.   0.   3.]
         [-18.   1.  -3.   0.  11.   0.]
         [  0.   0.   0.   3.   0.  85.]]
        Phase: TiNi
        Symmetry: monoclinic

        >>> from Elasticipy.tensors.elasticity import ComplianceTensor\n
        >>> ComplianceTensor.fromCrystalSymmetry(symmetry='monoclinic', diad='y', phase_name='TiNi',
        ...                                      S11=8, S12=-3, S13=-2,
        ...                                      S22=8, S23=-5, S33=10,
        ...                                      S44=12, S55=116, S66=12,
        ...                                      S15=14, S25=-8, S35=0, S46=0)
        Compliance tensor (in Voigt mapping):
        [[  8.  -3.  -2.   0.  14.   0.]
         [ -3.   8.  -5.   0.  -8.   0.]
         [ -2.  -5.  10.   0.   0.   0.]
         [  0.   0.   0.  12.   0.   0.]
         [ 14.  -8.   0.   0. 116.   0.]
         [  0.   0.   0.   0.   0.  12.]]
        Phase: TiNi
        Symmetry: monoclinic
        """
        warn('This function will be removed in a future release. Use {}.{}() instead'.format(cls.__name__,symmetry), DeprecationWarning, stacklevel=2)
        return cls._fromCrystalSymmetry(symmetry=symmetry, point_group=point_group, diad=diad, phase_name=phase_name,
                                       prefix=prefix, **kwargs)

    @classmethod
    def _fromCrystalSymmetry(cls, symmetry, phase_name, **kwargs):
        matrix = cls._matrixFromCrystalSymmetry(symmetry=symmetry, **kwargs)
        return cls(matrix, phase_name=phase_name, symmetry=symmetry)


    @classmethod
    def hexagonal(cls, *, C11=0., C12=0., C13=0., C33=0., C44=0., phase_name=None):
        """
        Create a fourth-order tensor from hexagonal symmetry.

        Parameters
        ----------
        C11, C12 , C13, C33, C44 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        transverse_isotropic : creates a transverse-isotropic tensor from engineering parameters
        cubic : create a tensor from cubic symmetry
        tetragonal : create a tensor from tetragonal symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='hexagonal', C11=C11, C12=C12, C13=C13, C33=C33, C44=C44,
                                       phase_name=phase_name, prefix='C')

    @classmethod
    def trigonal(cls, *, C11=0., C12=0., C13=0., C14=0., C33=0., C44=0., C15=0., phase_name=None):
        """
        Create a fourth-order tensor from trigonal symmetry.

        Parameters
        ----------
        C11, C12, C13, C14, C33, C44 : float
            Components of the tensor, using the Voigt notation
        C15 : float, optional
            C15 component of the tensor, only used for point groups 3 and -3.
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        tetragonal : create a tensor from tetragonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='trigonal', point_group='3',
                                        C11=C11, C12=C12, C13=C13, C14=C14, C15=C15,
                                        C33=C33, C44=C44, phase_name=phase_name, prefix='C')

    @classmethod
    def tetragonal(cls, *, C11=0., C12=0., C13=0., C33=0., C44=0., C16=0., C66=0., phase_name=None):
        """
        Create a fourth-order tensor from tetragonal symmetry.

        Parameters
        ----------
        C11,  C12, C13, C33, C44, C66 : float
            Components of the tensor, using the Voigt notation
        C16 : float, optional
            C16 component in Voigt notation (for point groups 4, -4 and 4/m only)
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        trigonal : create a tensor from trigonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='tetragonal', point_group='4',
                                        C11=C11, C12=C12, C13=C13, C16=C16,
                                        C33=C33, C44=C44, C66=C66, phase_name=phase_name, prefix='C')

    @classmethod
    def cubic(cls, *, C11=0., C12=0., C44=0., phase_name=None):
        """
        Create a fourth-order tensor from cubic symmetry.

        Parameters
        ----------
        C11 , C12, C44 : float
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        hexagonal : create a tensor from hexagonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='cubic', C11=C11, C12=C12, C44=C44, phase_name=phase_name, prefix='C')

    @classmethod
    def orthorhombic(cls, *, C11=0., C12=0., C13=0., C22=0., C23=0., C33=0., C44=0., C55=0., C66=0., phase_name=None):
        """
        Create a fourth-order tensor from orthorhombic symmetry.

        Parameters
        ----------
        C11, C12, C13, C22, C23, C33, C44, C55, C66 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='orthorhombic',
                                       C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55, C66=C66,
                                       phase_name=phase_name, prefix='C')

    @classmethod
    def monoclinic(cls, *, C11=0., C12=0., C13=0., C22=0., C23=0., C33=0., C44=0., C55=0., C66=0.,
                   C15=None, C25=None, C35=None, C46=None,
                   C16=None, C26=None, C36=None, C45=None,
                   phase_name=None):
        """
        Create a fourth-order tensor from monoclinic symmetry. It automatically detects whether the components are given
        according to the Y or Z diad, depending on the input arguments.

        For Diad || y, C15, C25, C35 and C46 must be provided.
        For Diad || z, C16, C26, C36 and C45 must be provided.

        Parameters
        ----------
        C11, C12 , C13, C22, C23, C33, C44, C55, C66 : float
            Components of the tensor, using the Voigt notation
        C15 : float, optional
            C15 component of the tensor (if Diad || y)
        C25 : float, optional
            C25 component of the tensor (if Diad || y)
        C35 : float, optional
            C35 component of the tensor (if Diad || y)
        C46 : float, optional
            C46 component of the tensor (if Diad || y)
        C16 : float, optional
            C16 component of the tensor (if Diad || z)
        C26 : float, optional
            C26 component of the tensor (if Diad || z)
        C36 : float, optional
            C36 component of the tensor (if Diad || z)
        C45 : float, optional
            C45 component of the tensor (if Diad || z)
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        triclinic : create a tensor from triclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        diad_y = not (None in (C15, C25, C35, C46))
        diad_z = not (None in (C16, C26, C36, C45))
        if diad_y and diad_z:
            raise KeyError('Ambiguous diad. Provide either C15, C25, C35 and C46; or C16, C26, C36 and C45')
        elif diad_y:
            return cls._fromCrystalSymmetry(symmetry='monoclinic', diad='y',
                                           C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55,
                                           C66=C66,
                                           C15=C15, C25=C25, C35=C35, C46=C46, phase_name=phase_name, prefix='C')
        elif diad_z:
            return cls._fromCrystalSymmetry(symmetry='monoclinic', diad='z',
                                           C11=C11, C12=C12, C13=C13, C22=C22, C23=C23, C33=C33, C44=C44, C55=C55,
                                           C66=C66,
                                           C16=C16, C26=C26, C36=C36, C45=C45, phase_name=phase_name, prefix='C')
        else:
            raise KeyError('For monoclinic symmetry, one should provide either C15, C25, C35 and C46, '
                           'or C16, C26, C36 and C45.')

    @classmethod
    def triclinic(cls, C11=0., C12=0., C13=0., C14=0., C15=0., C16=0.,
                  C22=0., C23=0., C24=0., C25=0., C26=0.,
                  C33=0., C34=0., C35=0., C36=0.,
                  C44=0., C45=0., C46=0.,
                  C55=0., C56=0.,
                  C66=0., phase_name=None):
        """

        Parameters
        ----------
        C11 , C12 , C13 , C14 , C15 , C16 , C22 , C23 , C24 , C25 , C26 , C33 , C34 , C35 , C36 , C44 , C45 , C46 , C55 , C56 , C66 : float
            Components of the tensor
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        matrix = np.array([[C11, C12, C13, C14, C15, C16],
                           [C12, C22, C23, C24, C25, C26],
                           [C13, C23, C33, C34, C35, C36],
                           [C14, C24, C34, C44, C45, C46],
                           [C15, C25, C35, C45, C55, C56],
                           [C16, C26, C36, C46, C56, C66]])
        return cls(matrix, phase_name=phase_name)

    def _single_tensor_only(self, fun_name=''):
        if self.ndim:
            err_msg = fun_name + ' is not suitable for tensor array. Consider subscripting (e.g. C[0].{}).'.format(fun_name)
            raise ValueError(err_msg)

    @property
    def Young_modulus(self):
        """
        Directional Young's modulus

        Returns
        -------
        SphericalFunction
            Young's modulus
        """
        self._single_tensor_only('Young_modulus')
        if isinstance(self, ComplianceTensor):
            S = self
        else:
            S = self.inv()
        def compute_young_modulus(u):
            a = np.einsum('ijkl,...i,...j,...k,...l->...', S.full_tensor(), u, u, u, u)
            return 1 / a

        return SphericalFunction(compute_young_modulus)

    @property
    def shear_modulus(self):
        """
        Directional shear modulus

        Returns
        -------
        HyperSphericalFunction
            Shear modulus
        """
        self._single_tensor_only('shear_modulus')
        if isinstance(self, ComplianceTensor):
            S = self
        else:
            S = self.inv()
        def compute_shear_modulus(u, v):
            G =  0.25/np.einsum('ijkl,...i,...j,...k,...l->...',S.full_tensor(),u,v,u,v)
            return G

        return HyperSphericalFunction(compute_shear_modulus)

    @property
    def Poisson_ratio(self):
        """
        Directional Poisson's ratio

        Returns
        -------
        HyperSphericalFunction
            Poisson's ratio

        Notes
        -----
        If the material undergoes tensile strain :math:`\\varepsilon_{ii}` along the i-th direction, the Poisson ratios
        are defined as:

        .. math::

            \\nu_{ij}=-\\frac{\\partial \\varepsilon_{jj}}{\\partial \\varepsilon_{ii}}

        where :math:`\\varepsilon_{jj}` denotes the (compressive) longitudinal strain along the j-th direction.
        """
        self._single_tensor_only('Poisson_ratio')
        if isinstance(self, ComplianceTensor):
            Sfull = self.full_tensor()
        else:
            Sfull = self.inv().full_tensor()
        def compute_PoissonRatio(u, v):
            numer = np.einsum('ijkl,...i,...j,...k,...l->...',Sfull,v,v,u,u)
            denom = np.einsum('ijkl,...i,...j,...k,...l->...',Sfull,u,u,u,u)
            return -numer / denom

        return HyperSphericalFunction(compute_PoissonRatio)

    @property
    def linear_compressibility(self):
        """
        Compute the directional linear compressibility.

        Returns
        -------
        SphericalFunction
            Directional linear compressibility

        See Also
        --------
        bulk_modulus : bulk modulus of the material
        """
        self._single_tensor_only('linear_compressibility')
        if isinstance(self, ComplianceTensor):
            S = self
        else:
            S = self.inv()
        def compute_linear_compressibility(u):
            return np.einsum('ijkk,...i,...j->...',S.full_tensor(),u,u)

        return SphericalFunction(compute_linear_compressibility)

    @property
    def bulk_modulus(self):
        """
        Compute the bulk modulus of the material

        Returns
        -------
        float or numpy.ndarray
            Bulk modulus

        See Also
        --------
        linear_compressibility : directional linear compressibility
        """
        return self.inv().bulk_modulus

    @property
    def lame1(self):
        """"
        Compute the first Lamé's parameter (only for isotropic materials).

        If the stiffness/compliance tensor is not isotropic, NaN is returned.

        Returns
        -------
        float
            First Lamé's parameter

        See Also
        --------
        lame2 : second Lamé's parameter
        """
        self._single_tensor_only('lame1')
        if self.is_isotropic():
            C11 = (self.C11 + self.C22 + self.C33) / 3
            return C11 - 2 * self.lame2
        else:
            return np.nan

    @property
    def lame2(self):
        """"
        Compute the second Lamé's parameter (only for isotropic materials).

        If the stiffness/compliance tensor is not isotropic, NaN is returned.

        Returns
        -------
        float
            Second Lamé's parameter

        See Also
        --------
        lame1 : first Lamé's parameter
        """
        self._single_tensor_only('lame2')
        if self.is_isotropic():
            return (self.C44 + self.C55 + self.C66) / 3
        else:
            return np.nan

    def Voigt_average(self, axis=None):
        """
        Compute the Voigt average of the stiffness tensor.

        If the tensor is a tensor array, all its values are considered. Otherwise (i.e. if single), the corresponding
        isotropic tensor is returned.

        Parameters
        ----------
        axis : int, optional
            If provided, the average is computed along this axis. Otherwise, the mean is computed on the flattened array.

        Returns
        -------
        StiffnessTensor
            Voigt average of stiffness tensor

        See Also
        --------
        Reuss_average : compute the Reuss average
        Hill_average : compute the Voigt-Reuss-Hill average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        if self.ndim:
            return self.mean(axis=axis)
        else:
            c = self._matrix
            A = c[0, 0] + c[1, 1] + c[2, 2]
            B = c[0, 1] + c[0, 2] + c[1, 2]
            C = c[3, 3] + c[4, 4] + c[5, 5]
            C11 = 1 / 5 *  A  + 2 / 15 * B + 4 / 15 * C
            C12 = 1 / 15 * A  + 4 / 15 * B - 2 / 15 * C
            C44 = (A - B) / 15 + C / 5
            mat = _isotropic_matrix(C11, C12, C44)
            return StiffnessTensor(mat, symmetry='isotropic', phase_name=self.phase_name)

    def Reuss_average(self, axis=None):
        """
        Compute the Reuss average of the stiffness tensor. If the tensor contains no orientation, we assume isotropic
        behaviour. Otherwise, the mean is computed over all orientations.

        Parameters
        ----------
        axis : int, optional
            If provided, axis to compute the average along with. If none, the average is computed on the flattened array

        Returns
        -------
        StiffnessTensor
            Reuss average of stiffness tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Hill_average : compute the Voigt-Reuss-Hill average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        return self.inv().Reuss_average(axis=axis).inv()

    def Hill_average(self, axis=None):
        """
        Compute the (Voigt-Reuss-)Hill average of the stiffness tensor. If the tensor contains no orientation, we assume
        isotropic behaviour. Otherwise, the mean is computed over all orientations.

        Parameters
        ----------
        axis : int, optional
            If provided, axis to compute the average along with. If none, the average is computed on the flattened array

        Returns
        -------
        StiffnessTensor
            Voigt-Reuss-Hill average of tensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Reuss_average : compute the Reuss average
        average : generic function for calling either the Voigt, Reuss or Hill average
        """
        Reuss = self.Reuss_average(axis=axis)
        Voigt = self.Voigt_average(axis=axis)
        return (Reuss + Voigt) * 0.5

    def average(self, method, axis=None):
        """
        Compute either the Voigt, Reuss, or Hill average of the stiffness tensor.

        This function is just a shortcut for Voigt_average(), Reuss_average(), or Hill_average() and Hill_average().

        Parameters
        ----------
        axis : int, optional
            If provided, axis to compute the average along with. If none, the average is computed on the flattened array
        method : str {'Voigt', 'Reuss', 'Hill'}
            Method to use to compute the average.

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        Voigt_average : compute the Voigt average
        Reuss_average : compute the Reuss average
        Hill_average : compute the Voigt-Reuss-Hill average
        """
        method = method.capitalize()
        if method in ('Voigt', 'Reuss', 'Hill'):
            fun = getattr(self, method + '_average')
            return fun(axis=axis)
        else:
            raise NotImplementedError('Only Voigt, Reus, and Hill are implemented.')

    @classmethod
    def isotropic(cls, E=None, nu=None, G=None, lame1=None, lame2=None, K=None, phase_name=None):
        """
        Create an isotropic stiffness tensor from two elasticity coefficients, namely: E, nu, G, lame1, or lame2.
        Exactly two of these coefficients must be provided. Note that lame2 is just an alias for G.

        Parameters
        ----------
        E : float, None
            Young modulus
        nu : float, None
            Poisson ratio
        G : float, None
            Shear modulus
        lame1 : float, None
            First Lamé coefficient
        lame2 : float, None
            Second Lamé coefficient (alias for G)
        K : float, None
            Bulk modulus
        phase_name : str, None
            Name to print

        Returns
        -------
            Corresponding isotropic stiffness tensor

        See Also
        --------
        transverse_isotropic : create a transverse-isotropic tensor

        Notes
        -----
        The units you use when passing the elastic moduli must be consistent with that of the stress tensor. For
        instance, if you expect to work in MPa, the Young's modulus and the Lamé's coefficient must be given in MPa as
        well.

        Examples
        --------
        On can check that the shear modulus for steel is around 82 GPa:

        >>> from Elasticipy.tensors.elasticity import StiffnessTensor
        >>> C=StiffnessTensor.isotropic(E=210e3, nu=0.28)
        >>> C.shear_modulus
        Hyperspherical function
        Min=82031.24999999991, Max=82031.25000000006
        """
        return ComplianceTensor.isotropic(E=E, nu=nu, G=G, lame1=lame1, lame2=lame2, K=K, phase_name=phase_name).inv()

    @classmethod
    def orthotropic(cls, *, Ex, Ey, Ez, Gxy, Gxz, Gyz,
                    nu_yx=None, nu_zx=None, nu_zy=None,
                    nu_xy=None, nu_xz=None, nu_yz=None, **kwargs):
        """
        Create a stiffness tensor corresponding to orthotropic symmetry, given the engineering constants.

        Exactly three Poisson ratios must be provided. See Notes for details.

        Parameters
        ----------
        Ex : float
            Young modulus along the x axis
        Ey : float
            Young modulus along the y axis
        Ez : float
            Young modulus along the z axis
        Gxy : float
            Shear modulus in the x-y plane
        Gxz : float
            Shear modulus in the x-z plane
        Gyz : float
            Shear modulus in the y-z plane
        nu_xy, nu_yx : float, optional
            Poisson ratio along x and y axes. Either nu_xy or nu_yx must be provided, not both.
        nu_xz, nu_zx : float, optional
            Poisson ratio along x and z axes. Either nu_xz or nu_zx must be provided, not both.
        nu_yz, nu_zy : float, optional
            Poisson ratio along y and z axes. Either nu_yz or nu_zy must be provided, not both.
        kwargs : dict, optional
            Keyword arguments to pass to the StiffnessTensor constructor

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        transverse_isotropic : create a stiffness tensor for transverse-isotropic symmetry

        Notes
        -----
        If the material undergoes tensile strain :math:`\\varepsilon_{ii}` along the i-th direction, the Poisson ratios
        are defined as:

        .. math::

            \\nu_{ij}=-\\frac{\\partial \\varepsilon_{jj}}{\\partial \\varepsilon_{ii}}

        where :math:`\\varepsilon_{jj}` denotes the (compressive) longitudinal strain along the j-th direction. If
        :math:`E_x` and :math:`E_y` are the Young moduli along `x` and `y`, we have:

        .. math::

            \\frac{\\nu_{xy}}{E_x} = \\frac{\\nu_{yx}}{E_y}

        """
        return ComplianceTensor.orthotropic(Ex=Ex, Ey=Ey, Ez=Ez, Gxy=Gxy, Gxz=Gxz, Gyz=Gyz,
                                            nu_yx=nu_yx, nu_zx=nu_zx, nu_zy=nu_zy,
                                            nu_xy=nu_xy, nu_xz=nu_xz, nu_yz=nu_yz).inv()

    @classmethod
    def transverse_isotropic(cls, *, Ex, Ez, Gxz, nu_yx=None, nu_xy=None, nu_zx=None, nu_xz=None, **kwargs):
        """
        Create a stiffness tensor corresponding to the transversely isotropic symmetry with respect to Z axis, given the
        engineering constants.

        Exactly two Poisson ratios must be provided (nu_xy or nu_yx, and nu_xz or nu_zx). See Notes for details.

        Parameters
        ----------
        Ex : float
            Young modulus along the x axis
        Ez : float
            Young modulus along the y axis
        Gxz : float
            Shear modulus in the x-z plane
        nu_xy, nu_yx : float, optional
            Poisson ratio along x and y. Either nu_xy or nu_yx must be provided, not both.
        nu_xz, nu_zx : float, optional
            Poisson ratio along x and z. Either nu_xz or nu_zx must be provided, not both.
        kwargs : dict
            Keyword arguments to pass to the StiffnessTensor constructor

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        orthotropic : create a stiffness tensor for orthotropic symmetry

        Notes
        -----
        If the material undergoes tensile strain :math:`\\varepsilon_{ii}` along the i-th direction, the Poisson ratios
        are defined as:

        .. math::

            \\nu_{ij}=-\\frac{\\partial \\varepsilon_{jj}}{\\partial \\varepsilon_{ii}}

        where :math:`\\varepsilon_{jj}` denotes the (compressive) longitudinal strain along the j-th direction. If
        :math:`E_x` and :math:`E_y` are the Young moduli along `x` and `y`, we have:

        .. math::

            \\frac{\\nu_{xy}}{E_x} = \\frac{\\nu_{yx}}{E_y}
        """
        nu_yx = _switch_poisson_ratios(nu_xy, nu_yx, Ex, Ex,'xy')
        Gxy = Ex / (2 * (1 + nu_yx))
        C = StiffnessTensor.orthotropic(Ex=Ex, Ey=Ex, Ez=Ez,
                                        nu_yx=nu_yx, nu_zx=nu_zx, nu_zy=nu_zx,
                                        nu_xz=nu_xz, nu_yz=nu_xz,
                                        Gxy=Gxy, Gxz=Gxz, Gyz=Gxz, **kwargs)
        C.symmetry = 'transverse-isotropic'
        return C

    def Christoffel_tensor(self, u):
        """
        Create the Christoffel tensor along a given direction, or set or directions.

        Parameters
        ----------
        u : list or np.ndarray
            3D direction(s) to compute the Christoffel tensor along with

        Returns
        -------
        Gamma : np.ndarray
            Array of Christoffel tensor(s). if u is a list of directions, Gamma[i] is the Christoffel tensor for
            direction  u[i].

        See Also
        --------
        wave_velocity : computes the p- and s-wave velocities.

        Notes
        -----
        For a given stiffness tensor **C** and a given unit vector **u**, the Christoffel tensor is defined as [2]_ :

            .. math:: M_{ij} = C_{iklj}.u_k.u_l

        """
        u_vec = np.atleast_2d(u)
        u_vec = (u_vec.T / np.linalg.norm(u_vec, axis=1)).T
        return np.einsum('inmj,pn,pm->pij', self.full_tensor(), u_vec, u_vec)

    def wave_velocity(self, rho):
        """
        Compute the wave velocities, given the mass density.

        Parameters
        ----------
        rho : float
            mass density. Its unit must be consistent with that of the stiffness tensor. See notes for hints.

        See Also
        --------
        Christoffel_tensor : Computes the Christoffel tensor along a given direction

        Returns
        -------
        c_p : SphericalFunction
            Velocity of the primary (compressive) wave
        c_s1 : SphericalFunction
            Velocity of the fast secondary (shear) wave
        c_s2 : SphericalFunction
            Velocity of the slow secondary (shear) wave

        Notes
        -----
        The estimation of the wave velocities is made by finding the eigenvalues of the Christoffel tensor [2]_.

        One should double-check the units. The table below provides hints about the unit you get, depending on the units
        you use for stiffness and the mass density:

        +-----------------+--------------+------------+-----------------------+
        | Stiffness       | Mass density | Velocities | Notes                 |
        +=================+==============+============+=======================+
        | Pa (N/m²)       | kg/m³        | m/s        | SI units              |
        +-----------------+--------------+------------+-----------------------+
        | GPa (10⁹ Pa)    | kg/dm³       | km/s       | Conversion factor     |
        +-----------------+--------------+------------+-----------------------+
        | GPa (10³ N/mm²) | kg/mm³       | m/s        | Consistent units      |
        +-----------------+--------------+------------+-----------------------+
        | MPa (10⁶ Pa)    | kg/m³        | km/s       | Conversion factor     |
        +-----------------+--------------+------------+-----------------------+
        | MPa (10³ N/mm²) | g/mm³        | m/s        | Consistent units      |
        +-----------------+--------------+------------+-----------------------+

        References
        ----------
        .. [2] J. W. Jaeken, S. Cottenier, Solving the Christoffel equation: Phase and group velocities, Computer Physics
               Communications (207), 2016, https://doi.org/10.1016/j.cpc.2016.06.014.

        """
        self._single_tensor_only('wave_velocity')
        def make_fun(index):
            def fun(n):
                Gamma = self.Christoffel_tensor(n)
                eig, _ = np.linalg.eigh(Gamma)
                eig_of_interest = eig[...,2-index]  # Switch ordering (descending order)
                return np.sqrt(eig_of_interest / rho)

            return fun

        return [SphericalFunction(make_fun(i)) for i in range(3)]

    @classmethod
    def from_MP(cls, ids, api_key=None):
        """
        Import stiffness tensor(s) from the Materials Project API, given their material ids.

        You need to register to `<https://materialsproject.org>`_ first to get an API key. This key can be explicitly
        passed as an argument (see below), or provided as an environment variable named MP_API_KEY.

        Parameters
        ----------
        ids : str or list of str
            ID(s) of the material to import (e.g. "mp-1048")
        api_key : str, optional
            API key to the Materials Project API. If not provided, it should be available as the API_KEY environment
            variable.

        Returns
        -------
        list of StiffnessTensor
            If one of the requested material ids was not found, the corresponding value in the list will be None.
        """
        try:
            from mp_api.client import MPRester
        except ImportError:
            raise ModuleNotFoundError('mp_api module is required for this function.')
        if type(ids) is str:
            Cdict = dict.fromkeys([ids])
        else:
            Cdict = dict.fromkeys(ids)
        with MPRester(api_key=api_key) as mpr:
            elasticity_doc = mpr.materials.elasticity.search(material_ids=ids)
            for material in elasticity_doc:
                key = str(material.material_id)
                if material.elastic_tensor is not None:
                    matrix = material.elastic_tensor.ieee_format
                    symmetry = material.symmetry.crystal_system.value
                    phase_name = material.formula_pretty
                    C = StiffnessTensor(np.asarray(matrix), symmetry=str(symmetry), phase_name=phase_name)
                else:
                    C = None
                Cdict[key] = C
            if elasticity_doc:
                if isinstance(ids, str):
                    return C
                else:
                    return [Cdict[id] for id in ids]
            else:
                return None

    @classmethod
    def weighted_average(cls, Cs, volume_fractions, method):
        """
        Compute the weighted average of a list of stiffness tensors, with respect to a given method (Voigt, Reuss or
        Hill).

        Parameters
        ----------
        Cs : list of StiffnessTensor or list of ComplianceTensor or tuple of StiffnessTensor or tuple of ComplianceTensor
            Series of tensors to compute the average from
        volume_fractions : iterable of floats
            Volume fractions of each phase
        method : str, {'Voigt', 'Reuss', 'Hill'}
            Method to use. It can be 'Voigt', 'Reuss', or 'Hill'.

        Returns
        -------
        StiffnessTensor
            Average tensor
        """
        if np.all([isinstance(a, ComplianceTensor) for a in Cs]):
            Cs = [C.inv() for C in Cs]
        if np.all([isinstance(a, StiffnessTensor) for a in Cs]):
            C_stack = np.array([C._matrix for C in Cs])
            method = method.capitalize()
            if method == 'Voigt':
                C_avg = np.average(C_stack, weights=volume_fractions, axis=0)
                return StiffnessTensor(C_avg)
            elif method == 'Reuss':
                S_stack = np.linalg.inv(C_stack)
                S_avg = np.average(S_stack, weights=volume_fractions, axis=0)
                return StiffnessTensor(np.linalg.inv(S_avg))
            elif method == 'Hill':
                C_voigt = cls.weighted_average(Cs, volume_fractions, 'Voigt')
                C_reuss = cls.weighted_average(Cs, volume_fractions, 'Reuss')
                return (C_voigt + C_reuss) * 0.5
            else:
                raise ValueError('Method must be either Voigt, Reuss or Hill.')
        else:
            raise ValueError('The first argument must be either a list of ComplianceTensors or '
                             'a list of StiffnessTensor.')

    @property
    def universal_anisotropy(self):
        """
        Compute the universal anisotropy factor.

        The larger the value, the more likely the material will behave in an anisotropic way.

        Returns
        -------
        float
            The universal anisotropy factor.

        Notes
        -----
        The universal anisotropy factor is defined as [3]_:

        .. math::

            5\\frac{G_v}{G_r} + \\frac{K_v}{K_r} - 6

        References
        ----------
        .. [3] S. I. Ranganathan and M. Ostoja-Starzewski, Universal Elastic Anisotropy Index,
           *Phys. Rev. Lett.*, 101(5), 055504, 2008. https://doi.org/10.1103/PhysRevLett.101.055504
        """
        self._single_tensor_only('universal_anisotropy')
        Cvoigt = self.Voigt_average()
        Creuss = self.Reuss_average()
        Gv = Cvoigt._matrix[3, 3]
        Gr = Creuss._matrix[3, 3]
        Kv = Cvoigt.bulk_modulus
        Kr = Creuss.bulk_modulus
        return 5 * Gv / Gr + Kv / Kr - 6

    def Zener_ratio(self, tol=1e-4):
        """
        Compute the Zener ratio (Z). Only valid for cubic symmetry.

        This function first checks that the tensor has cubic symmetry within a given tolerance. If not, an error is
        raised.

        Parameters
        ----------
        tol : float, optional
            Tolerance to consider that the material has cubic symmetry

        Returns
        -------
        float
            Zener ratio (NaN is the symmetry is not cubic)

        Notes
        -----
        If the tensor is written in canonical base with Voigt mapping, the Zener ratio is defined as:

        .. math::

                Z=\\frac{ 2C_{44} }{C_{11} - C_{12}}

        The present implementation takes advantage of eigenstiffness to compute the Zener ratio in *any* base, i.e. even
        if the tensor is not given in canonical base (e.g. if rotated).

        See Also
        --------
        universal_anisotropy : compute the universal anisotropy factor
        eig_stiffnesses : eigenstiffnesses of the tensor
        is_cubic : check whether the tensor has cubic symmetry or not

        Examples
        --------
        >>> from Elasticipy.tensors.elasticity import StiffnessTensor
        >>> C = StiffnessTensor.cubic(C11=200, C12=40, C44=20)
        >>> C.Zener_ratio()
        0.25

        which obvisouly corresponds to 2.C44/(C11-C12).

        Now, rotate the tensor and see how it looks like:

        >>> from scipy.spatial.transform import Rotation
        >>> g = Rotation.from_euler('Z', 30, degrees=True)
        >>> C_rot = C*g
        >>> C_rot
        Stiffness tensor (in Voigt mapping):
        [[155.          85.          40.           0.           0.
           25.98076211]
         [ 85.         155.          40.           0.           0.
          -25.98076211]
         [ 40.          40.         200.           0.           0.
            0.        ]
         [  0.           0.           0.          20.           0.
            0.        ]
         [  0.           0.           0.           0.          20.
            0.        ]
         [ 25.98076211 -25.98076211   0.           0.           0.
           65.        ]]
        Symmetry: cubic

        Still, we have
        >>> C_rot.Zener_ratio()
        0.24999999999999983
        """
        if self.is_isotropic():
            return 1.0
        elif self.is_cubic():
            eigs, orders = self.eig_stiffnesses_multiplicity(1e-4)
            numer = eigs[orders==3][0]  # 2*C44
            denom = eigs[orders==2][0]  # C11-C12
            return numer / denom
        else:
            raise ValueError('The tensor does not seem to have cubic symmetry within the given tolerance ({})'.format(tol))

    def to_pymatgen(self):
        """
        Convert the stiffness tensor (from Elasticipy) to Python Materials Genomics (Pymatgen) format.

        Returns
        -------
        pymatgen.analysis.elasticity.elastic.ElasticTensor
            Stiffness tensor for pymatgen
        """
        try:
            from pymatgen.analysis.elasticity import elastic as matgenElast
        except ImportError:
            raise ModuleNotFoundError('pymatgen module is required for this function.')
        return matgenElast.ElasticTensor(self.full_tensor())

    def to_Kelvin(self):
        """
        Returns all the tensor components using the Kelvin(-Mandel) mapping convention.

        Returns
        -------
        numpy.ndarray
            (6,6) matrix, according to the Kelvin mapping

        See Also
        --------
        eig : returns the eigenvalues and the eigenvectors of the Kelvin's matrix
        from_Kelvin : Construct a fourth-order tensor from its (6,6) Kelvin matrix

        Notes
        -----
        This mapping convention is defined as follows [4]_:

        .. math::

            C_K = \\begin{bmatrix}
                C_{11}          & C_{12}            & C_{13}            & \\sqrt{2}C_{14} & \\sqrt{2}C_{15} & \\sqrt{2}C_{16}\\\\
                C_{12}          & C_{22}            & C_{23}            & \\sqrt{2}C_{24} & \\sqrt{2}C_{25} & \\sqrt{2}C_{26}\\\\
                C_{13}          & C_{23}            & C_{33}            & \\sqrt{2}C_{34} & \\sqrt{2}C_{35} & \\sqrt{2}C_{36}\\\\
                \\sqrt{2}C_{14} & \\sqrt{2}C_{24}   & \\sqrt{2}C_{34}   & 2C_{44}         & 2C_{45}         & 2C_{46}\\\\
                \\sqrt{2}C_{15} & \\sqrt{2}C_{25}   & \\sqrt{2}C_{35}   & 2C_{45}         & 2C_{55}         & 2C_{56}\\\\
                \\sqrt{2}C_{16} & \\sqrt{2}C_{26}   & \\sqrt{2}C_{36}   & 2C_{46}         & 2C_{56}         & 2C_{66}\\\\
            \\end{bmatrix}


        References
        ----------
        .. [4] Helbig, K. (2013). What Kelvin might have written about Elasticity. Geophysical Prospecting, 61(1), 1-20.
            doi: 10.1111/j.1365-2478.2011.01049.x
        """
        kelvin_mapping = KelvinMapping()
        return self._matrix /self.mapping.matrix * kelvin_mapping.matrix

    def eig(self):
        """
        Compute the eigenstiffnesses and the eigenstrains.

        Solve the eigenvalue problem from the Kelvin matrix of the stiffness tensor (see Notes).

        Returns
        -------
        numpy.ndarray
            Array of 6 eigenstiffnesses (eigenvalues of the stiffness matrix)
        numpy.ndarray
            (6,6) array of eigenstrains (eigenvectors of the stiffness matrix)

        See Also
        --------
        to_Kelvin : returns the stiffness components as a (6,6) matrix, according to the Kelvin mapping convention.
        eig_stiffnesses : returns the eigenstiffnesses only
        eig_strains : returns the eigenstrains only

        Notes
        -----
        The definition for eigenstiffnesses and the eigenstrains are introduced in [4]_.
        """
        return np.linalg.eigh(self.to_Kelvin())

    @property
    def eig_stiffnesses(self):
        """
        Compute the eigenstiffnesses given by the Kelvin's matrix for stiffness.

        Returns
        -------
        numpy.ndarray
            6 eigenvalues of the Kelvin's stiffness matrix, in ascending order

        See Also
        --------
        eig : returns the eigenstiffnesses and the eigenstrains
        eig_strains : returns the eigenstrains only
        eig_stiffnesses_multiplicity : returns the unique values of eigenstiffnesses with multiplicity
        """
        return np.linalg.eigvalsh(self.to_Kelvin())

    @property
    def eig_strains(self):
        """
        Compute the eigenstrains from the Kelvin's matrix for stiffness

        Returns
        -------
        numpy.ndarray
            (6,6) matrix of eigenstrains, sorted by ascending order of eigenstiffnesses.

        See Also
        --------
        eig : returns both the eigenvalues and the eigenvectors of the Kelvin matrix
        """
        return self.eig()[1]

    @property
    def eig_compliances(self):
        """
        Compute the eigencompliances from the Kelvin's matrix of stiffness

        Returns
        -------
        numpy.ndarray
            Inverses of the 6 eigenvalues of the Kelvin's stiffness matrix, in descending order

        See Also
        --------
        eig_stiffnesses : compute the eigenstiffnesses from the Kelvin's matrix of stiffness
        """
        return 1/self.eig_stiffnesses

    @classmethod
    def from_Kelvin(cls, matrix, **kwargs):
        """
        Create a tensor from the (6,6) matrix following the Kelvin(-Mandel) mapping convention

        Parameters
        ----------
        matrix : list or numpy.ndarray
            (6,6) matrix of components
        kwargs
            keyword arguments passed to the constructor

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        to_Kelvin : return the components as a (6,6) matrix following the Kelvin convention
        """
        kelvin_mapping = KelvinMapping()
        t = cls(matrix / kelvin_mapping.matrix, **kwargs)
        t._matrix *= t.mapping.matrix
        return t

    def eig_stiffnesses_multiplicity(self, tol=1e-4):
        """
        Compute the eigenstiffnesses, then returns the multiplicity of each eigenstiffness.

        Given an absolute tolerance, duplicates in eigenstiffnesses are considered to compute the multiplicity of each
        value.

        Parameters
        ----------
        tol : float, optional
            Absolute tolerance to assume that two distinct eigenstiffnesses are the same

        Returns
        -------
        numpy.ndarray
            Unique values of eigenstiffnesses, sorted by increasing multiplicity
        numpy.ndarray
            Multiplicity of each unique eigenstiffness, sorted by ascending value

        See Also
        --------
        eig_stiffnesses : compute the eigenstiffnesses

        Examples
        --------
        >>> from Elasticipy.tensors.elasticity import StiffnessTensor
        >>> C = StiffnessTensor.cubic(C11=186, C12=134, C44=77)
        >>> C.eig_stiffnesses
        array([ 52.,  52., 154., 154., 154., 454.])
        >>> C.eig_stiffnesses_multiplicity()
        (array([ 52., 154., 454.]), array([2, 3, 1]))
        """
        eig = self.eig_stiffnesses
        counts = []
        uniques = []
        while eig.size:
            duplicates = np.isclose(eig[0], eig, atol=tol)
            counts.append(np.count_nonzero(duplicates))
            uniques.append(eig[0])
            eig = eig[np.logical_not(duplicates)]
        return np.array(uniques), np.array(counts)

    def _check_eig_signature(self, signature, tol):
        if self.shape:
            flat = self.flatten()
            bool_array = np.zeros(flat.shape[0], dtype=bool)
            for i, t in enumerate(flat):
                _, order = t.eig_stiffnesses_multiplicity(tol=tol)
                bool_array[i] = np.array_equal(np.sort(order), np.sort(signature))
            return bool_array.reshape(self.shape)
        else:
            _, order = self.eig_stiffnesses_multiplicity(tol=tol)
            return np.array_equal(np.sort(order), np.sort(signature))

    def is_isotropic(self, tol=0.01):
        """Check that the tensor corresponds to isotropic symmetry, within a given tolerance.

        The method relies on the multiplicity of eigenstiffnesses

        Parameters
        ----------
        tol : float
            Absolute tolerance to consider multiplicity of eigenstiffnesses

        Returns
        -------
        bool or numpy.ndarray
            If the tensor is single, the returned value is boolean.
            If the object is a tensor array, the returned value is an array of bools, the same shape as the tensor array.

        See Also
        --------
        is_cubic : check if the stiffness tensor has cubic symmetry
        is_tetragonal : check if the stiffness tensor has tetragonal symmetry
        eig_stiffnesses : compute eigenstiffnesses
        """
        return self._check_eig_signature([1, 5], tol)

    def is_cubic(self, tol=0.01):
        """Check that the tensor corresponds to cubic symmetry, within a given tolerance.

        The method relies on the multiplicity of eigenstiffnesses.

        Parameters
        ----------
        tol : float, optional
            Absolute tolerance to consider multiplicity of eigenstiffnesses

        Returns
        -------
        bool or numpy.ndarray
            If the tensor is single, the returned value is boolean.
            If the object is a tensor array, the returned value is an array of bools, the same shape as the tensor array.

        See Also
        --------
        is_isotropic : check if the stiffness tensor is isotropic
        is_tetragonal : check if the stiffness tensor has tetragonal symmetry
        eig_stiffnesses : compute eigenstiffnesses

        Examples
        --------
        >>> from Elasticipy.tensors.elasticity import StiffnessTensor
        >>> from scipy.spatial.transform import Rotation
        >>> C = StiffnessTensor.cubic(C11=186, C12=134, C44=77)
        >>> C_rotated = C * Rotation.random(random_state=123)
        >>> C_rotated
        Stiffness tensor (in Voigt mapping):
        [[237.71171578  96.41409344 119.87419078   8.1901353   -3.63846312
          -20.34233446]
         [ 96.41409344 250.74909842 106.83680814   9.33462785  -6.52548033
            0.99714278]
         [119.87419078 106.83680814 227.28900108 -17.52476315  10.16394345
           19.34519167]
         [  8.1901353    9.33462785 -17.52476315  49.83680814  19.34519167
           -6.52548033]
         [ -3.63846312  -6.52548033  10.16394345  19.34519167  62.87419078
            8.1901353 ]
         [-20.34233446   0.99714278  19.34519167  -6.52548033   8.1901353
           39.41409344]]
        Symmetry: cubic

        Once rotated, it is not clear if the stiffness tensors has cubic symmetry. Yet:
        >>> C_rotated.is_cubic()
        True

        """
        return np.logical_or(self._check_eig_signature([1, 2, 3], tol), self.is_isotropic(tol=tol))

    def is_tetragonal(self, tol=0.01):
        """Check that the tensor corresponds to tetragonal symmetry, within a given tolerance.

        The method relies on the multiplicity of eigenstiffnesses.

        Parameters
        ----------
        tol : float
            Absolute tolerance to consider multiplicity of eigenstiffnesses

        Returns
        -------
        bool or numpy.ndarray
            If the tensor is single, the returned value is boolean.
            If the object is a tensor array, the returned value is an array of bools, the same shape as the tensor array.

        See Also
        --------
        is_isotropic : check if the stiffness tensor is isotropic
        is_cubic : check if the stiffness tensor has cubic symmetry
        eig_stiffnesses : compute eigenstiffnesses
        """
        return np.logical_or(self._check_eig_signature([1, 1, 1, 1, 2], tol), self.is_cubic(tol=tol))


class ComplianceTensor(StiffnessTensor):
    """
    Class for manipulating compliance tensors
    """
    tensor_name = 'Compliance'
    _C11_C12_factor = 2.0
    _component_prefix = 'S'
    _C46_C56_factor = 2.0

    def __init__(self, C, check_positive_definite=True, mapping=VoigtMapping(tensor='Compliance'), **kwargs):
        super().__init__(C, check_positive_definite=check_positive_definite, mapping=mapping, **kwargs)
        self.mapping_name = 'Voigt'

    def __mul__(self, other):
        if isinstance(other, StressTensor):
            new_tensor = self.ddot(other)
            return StrainTensor(new_tensor.matrix)
        elif isinstance(other, StrainTensor):
            raise ValueError('You cannot multiply a compliance tensor with Strain tensor.')
        else:
            return super().__mul__(other)

    def inv(self):
        """
        Compute the reciprocal stiffness tensor

        Returns
        -------
        StiffnessTensor
            Reciprocal tensor
        """
        S = np.linalg.inv(self._matrix)
        return StiffnessTensor(S, symmetry=self.symmetry, phase_name=self.phase_name)

    def Reuss_average(self, axis=None):
        if self.ndim:
            return self.mean(axis=axis)
        else:
            s = self._matrix
            A = s[0, 0] + s[1, 1] + s[2, 2]
            B = s[0, 1] + s[0, 2] + s[1, 2]
            C = s[3, 3] + s[4, 4] + s[5, 5]
            S11 = 1 / 5 *  A + 2 / 15 * B + 1 / 15 * C
            S12 = 1 / 15 * A + 4 / 15 * B - 1 / 30 * C
            S44 = 4 / 15 * (A - B) + 1 / 5 * C
            mat = _isotropic_matrix(S11, S12, S44)
            return ComplianceTensor(mat, symmetry='isotropic', phase_name=self.phase_name)

    def Voigt_average(self, axis=None):
        return self.inv().Voigt_average(axis=axis).inv()

    def Hill_average(self, axis=None):
        return self.inv().Hill_average(axis=axis).inv()

    @classmethod
    def isotropic(cls, E=None, nu=None, G=None, lame1=None, lame2=None, K=None, phase_name=None):
        if lame2 is not None:
            if G is None:
                G = lame2
            else:
                raise ValueError('G and lame2 cannot be provided together.')
        n_specified = sum(v is not None for v in [E, nu, lame1, G, K])
        if n_specified != 2:
            raise ValueError("Exactly two values are required among E, nu, G, K, lame1 and lame2.")
        if K is not None:
            if E is not None:
                G = 3 * K * E / (9 * K - E)
                nu = (3 * K - E) / 6 / K
            elif lame1 is not None:
                E = 9 * K * (K - lame1) / (3 * K -lame1)
                G= 3 * (K - lame1) / 2
                nu = lame1 / (3*K-lame1)
            elif G is not None:
                E = 9 * K * G / (3 * K + G)
                nu = (3 * K - 2 * G) / 2 / (3 * K + G)
            elif nu is not None:
                E = 3 * K * (1 - 2 * nu)
                G = E / 2 / (1 + nu)
        elif E is not None:
            if lame1 is not None:
                R = np.sqrt(E**2 + 9*lame1**2+2*E*lame1)
                G = (E - 3 * lame1 + R) / 4
                nu = 2 * lame1 / (E + lame1 + R)
            elif G is not None:
                nu = E / 2 / G - 1
            elif nu is not None:
                G = E / 2 / (1 + nu)
        elif lame1 is not None:
            if G is not None:
                E = G * (3 * lame1 + 2 * G) / (lame1 + G)
                nu = lame1 / 2 / (lame1 + G)
            elif nu is not None:
                E = lame1 * ( 1 + nu) * (1 - 2 * nu) / nu
                G = lame1 * (1 - 2 * nu) / 2 / nu
        elif (nu is not None) and (G is not None):
            E = 2 * G * (1 + nu)
        S11 = 1/E
        S12 = -nu/E
        S44 = 1 / G
        S_mat = _isotropic_matrix(S11, S12, S44)
        return ComplianceTensor(S_mat, symmetry='isotropic', phase_name=phase_name)

    @classmethod
    def orthotropic(cls, *, Ex, Ey, Ez, Gxy, Gxz, Gyz,
                    nu_yx=None, nu_zx=None, nu_zy=None,
                    nu_xy=None, nu_xz=None, nu_yz=None, **kwargs):
        nu_yx = _switch_poisson_ratios(nu_xy, nu_yx, Ex, Ey,'xy')
        nu_zx = _switch_poisson_ratios(nu_xz, nu_zx, Ex, Ez,'xz')
        nu_zy = _switch_poisson_ratios(nu_yz, nu_zy, Ey, Ez,'yz')
        tri_sup = np.array([[1 / Ex, -nu_yx / Ey, -nu_zx / Ez, 0, 0, 0],
                            [0, 1 / Ey, -nu_zy / Ez, 0, 0, 0],
                            [0, 0, 1 / Ez, 0, 0, 0],
                            [0, 0, 0, 1 / Gyz, 0, 0],
                            [0, 0, 0, 0, 1 / Gxz, 0],
                            [0, 0, 0, 0, 0, 1 / Gxy]])
        S = tri_sup + np.tril(tri_sup.T, -1)
        return ComplianceTensor(S, symmetry='orthotropic', **kwargs)

    @classmethod
    def transverse_isotropic(cls, *args, **kwargs):
        return super().transverse_isotropic(*args, **kwargs).inv()

    @classmethod
    def weighted_average(cls, *args):
        return super().weighted_average(*args).inv()

    @property
    def bulk_modulus(self):
        matrix_t = self._matrix.T
        sub_matrix = matrix_t[0:3, 0:3]
        return 1 / np.sum(sub_matrix, axis=(0,1))

    @property
    def universal_anisotropy(self):
        """
        Compute the universal anisotropy factor.

        It is actually an alias for inv().universal_anisotropy.

        Returns
        -------
        float
            Universal anisotropy factor
        """
        return self.inv().universal_anisotropy

    def to_pymatgen(self):
        """
        Convert the compliance tensor (from Elasticipy) to Python Materials Genomics (Pymatgen) format.

        Returns
        -------
        ComplianceTensor
            Compliance tensor for pymatgen
        """
        try:
            from pymatgen.analysis.elasticity import elastic as matgenElast
        except ImportError:
            raise ModuleNotFoundError('pymatgen module is required for this function.')
        return matgenElast.ComplianceTensor(self.full_tensor())

    def eig(self):
        """
        Compute the eigencompliances and the eigenstresses.

        Solve the eigenvalue problem from the Kelvin matrix of the compliance tensor (see Notes).

        Returns
        -------
        numpy.ndarray
            Array of 6 eigencompliances (eigenvalues of the stiffness matrix)
        numpy.ndarray
            (6,6) array of eigenstresses (eigenvectors of the stiffness matrix)

        See Also
        --------
        Kelvin : returns the stiffness components as a (6,6) matrix, according to the Kelvin mapping convention.
        eig_compliances : returns the eigencompliances only
        eig_stresses : returns the eigenstresses only

        Notes
        -----
        The definition for eigencompliances and the eigenstresses are introduced in [4]_.
        """
        return np.linalg.eigh(self.to_Kelvin())

    @property
    def eig_compliances(self):
        """
        Compute the eigencompliances given by the Kelvin's matrix for stiffness.

        Returns
        -------
        numpy.ndarray
            6 eigenvalues of the Kelvin's compliance matrix, in ascending order

        See Also
        --------
        eig : returns the eigencompliances and the eigenstresses
        eig_strains : returns the eigenstresses only
        """
        return np.linalg.eigvalsh(self.to_Kelvin())

    @property
    def eig_stresses(self):
        """
        Compute the eigenstresses from the Kelvin's matrix for stiffness

        Returns
        -------
        numpy.ndarray
            (6,6) matrix of eigenstresses, sorted by ascending order of eigencompliances.

        See Also
        --------
        eig : returns both the eigencompliances and the eigenstresses
        """
        return self.eig()[1]

    @property
    def eig_stiffnesses(self):
        """
        Compute the eigenstiffnesses from the Kelvin's matrix of compliance

        Returns
        -------
        numpy.ndarray
            inverses of 6 eigenvalues of the Kelvin's compliance matrix, in descending order

        See Also
        --------
        eig_compliances : compute the eigencompliances from the Kelvin's matrix of compliance
        """
        return 1/self.eig_compliances

    @property
    def lame1(self):
        return self.inv().lame1

    @property
    def lame2(self):
        return self.inv().lame2


    @classmethod
    def hexagonal(cls, *, S11=0., S12=0., S13=0., S33=0., S44=0., phase_name=None):
        """
        Create a fourth-order tensor from hexagonal symmetry.

        Parameters
        ----------
        S11, S12 , S13, S33, S44 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        transverse_isotropic : creates a transverse-isotropic tensor from engineering parameters
        cubic : create a tensor from cubic symmetry
        tetragonal : create a tensor from tetragonal symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='hexagonal', S11=S11, S12=S12, S13=S13, S33=S33, S44=S44,
                                       phase_name=phase_name, prefix='S')

    @classmethod
    def trigonal(cls, *, S11=0., S12=0., S13=0., S14=0., S33=0., S44=0., S15=0., phase_name=None):
        """
        Create a fourth-order tensor from trigonal symmetry.

        Parameters
        ----------
        S11, S12, S13, S14, S33, S44 : float
            Components of the tensor, using the Voigt notation
        S15 : float, optional
            S15 component of the tensor, only used for point groups 3 and -3.
        phase_name : str, optional
            Phase name to display
        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        tetragonal : create a tensor from tetragonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='trigonal', point_group='3',
                                        S11=S11, S12=S12, S13=S13, S14=S14, S15=S15,
                                        S33=S33, S44=S44, phase_name=phase_name, prefix='S')

    @classmethod
    def tetragonal(cls, *, S11=0., S12=0., S13=0., S33=0., S44=0., S16=0., S66=0., phase_name=None):
        """
        Create a fourth-order tensor from tetragonal symmetry.

        Parameters
        ----------
        S11,  S12, S13, S33, S44, S66 : float
            Components of the tensor, using the Voigt notation
        S16 : float, optional
            S16 component in Voigt notation (for point groups 4, -4 and 4/m only)
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        trigonal : create a tensor from trigonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='tetragonal', point_group='4',
                                        S11=S11, S12=S12, S13=S13, S16=S16,
                                        S33=S33, S44=S44, S66=S66, phase_name=phase_name, prefix='S')

    @classmethod
    def cubic(cls, *, S11=0., S12=0., S44=0., phase_name=None):
        """
        Create a fourth-order tensor from cubic symmetry.

        Parameters
        ----------
        S11 , S12, S44 : float
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        StiffnessTensor

        See Also
        --------
        hexagonal : create a tensor from hexagonal symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='cubic', S11=S11, S12=S12, S44=S44, phase_name=phase_name, prefix='S')

    @classmethod
    def orthorhombic(cls, *, S11=0., S12=0., S13=0., S22=0., S23=0., S33=0., S44=0., S55=0., S66=0., phase_name=None):
        """
        Create a fourth-order tensor from orthorhombic symmetry.

        Parameters
        ----------
        S11, S12, S13, S22, S23, S33, S44, S55, S66 : float
            Components of the tensor, using the Voigt notation
        phase_name : str, optional
            Phase name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        return cls._fromCrystalSymmetry(symmetry='orthorhombic',
                                       S11=S11, S12=S12, S13=S13, S22=S22, S23=S23, S33=S33, S44=S44, S55=S55, S66=S66,
                                       phase_name=phase_name, prefix='S')

    @classmethod
    def monoclinic(cls, *, S11=0., S12=0., S13=0., S22=0., S23=0., S33=0., S44=0., S55=0., S66=0.,
                   S15=None, S25=None, S35=None, S46=None,
                   S16=None, S26=None, S36=None, S45=None,
                   phase_name=None):
        """
        Create a fourth-order tensor from monoclinic symmetry. It automatically detects whether the components are given
        according to the Y or Z diad, depending on the input arguments.

        For Diad || y, S15, S25, S35 and S46 must be provided.
        For Diad || z, S16, S26, S36 and S45 must be provided.

        Parameters
        ----------
        S11, S12 , S13, S22, S23, S33, S44, S55, S66 : float
            Components of the tensor, using the Voigt notation
        S15 : float, optional
            S15 component of the tensor (if Diad || y)
        S25 : float, optional
            S25 component of the tensor (if Diad || y)
        S35 : float, optional
            S35 component of the tensor (if Diad || y)
        S46 : float, optional
            S46 component of the tensor (if Diad || y)
        S16 : float, optional
            S16 component of the tensor (if Diad || z)
        S26 : float, optional
            S26 component of the tensor (if Diad || z)
        S36 : float, optional
            S36 component of the tensor (if Diad || z)
        S45 : float, optional
            S45 component of the tensor (if Diad || z)
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        triclinic : create a tensor from triclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        diad_y = not (None in (S15, S25, S35, S46))
        diad_z = not (None in (S16, S26, S36, S45))
        if diad_y and diad_z:
            raise KeyError('Ambiguous diad. Provide either S15, S25, S35 and S46; or S16, S26, S36 and S45')
        elif diad_y:
            return cls._fromCrystalSymmetry(symmetry='monoclinic', diad='y',
                                           S11=S11, S12=S12, S13=S13, S22=S22, S23=S23, S33=S33, S44=S44, S55=S55,
                                           S66=S66,
                                           S15=S15, S25=S25, S35=S35, S46=S46, phase_name=phase_name, prefix='S')
        elif diad_z:
            return cls._fromCrystalSymmetry(symmetry='monoclinic', diad='z',
                                           S11=S11, S12=S12, S13=S13, S22=S22, S23=S23, S33=S33, S44=S44, S55=S55,
                                           S66=S66,
                                           S16=S16, S26=S26, S36=S36, S45=S45, phase_name=phase_name, prefix='S')
        else:
            raise KeyError('For monoclinic symmetry, one should provide either S15, S25, S35 and S46, '
                           'or S16, S26, S36 and S45.')

    @classmethod
    def triclinic(cls, S11=0., S12=0., S13=0., S14=0., S15=0., S16=0.,
                  S22=0., S23=0., C24=0., S25=0., S26=0.,
                  S33=0., C34=0., S35=0., S36=0.,
                  S44=0., S45=0., S46=0.,
                  S55=0., C56=0.,
                  S66=0., phase_name=None):
        """

        Parameters
        ----------
        S11 , S12 , S13 , S14 , S15 , S16 , S22 , S23 , C24 , S25 , S26 , S33 , C34 , S35 , S36 , S44 , S45 , S46 , S55 , C56 , S66 : float
            Components of the tensor
        phase_name : str, optional
            Name to display

        Returns
        -------
        FourthOrderTensor

        See Also
        --------
        monoclinic : create a tensor from monoclinic symmetry
        orthorhombic : create a tensor from orthorhombic symmetry
        """
        matrix = np.array([[S11, S12, S13, S14, S15, S16],
                           [S12, S22, S23, C24, S25, S26],
                           [S13, S23, S33, C34, S35, S36],
                           [S14, C24, C34, S44, S45, S46],
                           [S15, S25, S35, S45, S55, C56],
                           [S16, S26, S36, S46, C56, S66]])
        return cls(matrix, phase_name=phase_name)


from Elasticipy.tensors.elasticity import StiffnessTensor
from scipy.spatial.transform import Rotation

################################################
# The lines below are only required in PyCharm
################################################
import matplotlib as mpl
mpl.use('Qt5Agg')
from qtpy import QtWidgets
app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])
################################################

# First, let consider the NiTi material:
C = StiffnessTensor.monoclinic(phase_name='TiNi',
                                        C11=231, C12=127, C13=104,
                                        C22=240, C23=131, C33=175,
                                        C44=81, C55=11, C66=85,
                                        C15=-18, C25=1, C35=-3, C46=3)
print(C)

# Let's have a look on its Young modulus
E = C.Young_modulus
# See min/max values
print(E)

# Now illustrate the spatial dependence
fig,_ = E.plot_xyz_sections()   # As 2D sections...
fig.show()
fig,_ = E.plot3D()              # ...or in 3D
fig.show()
fig,_ =E.plot_as_pole_figure()  # or even with PF
fig.show()

# Print out the maximum value for the Young modulus
print(E.max())

# Apply a random rotation on stiffness tensor
rotation = Rotation.from_euler('ZXZ', [90, 45, 0], degrees=True)
Crot = C*rotation
# Check that the Young modulus has changed as well
fig, _ = Crot.Young_modulus.plot3D()
fig.show()

# Now let's consider the shear modulus
G = C.shear_modulus
fig,_ = G.plot_xyz_sections()   # Plot sections with min, max and mean
fig.show()

fig,_ = G.plot3D()              # And plot it in 3D
fig.show()
print(G.min())
print(G.max())

# Finally, let's have a look on the Poisson ratio
nu = C.Poisson_ratio
fig,_ = nu.plot_xyz_sections()
fig.show()
fig,_ = nu.plot3D(which='max')
fig.show()

# Check out the maximum and minimum values
print(nu.min())
print(nu.max())

# Now let consider a finite set of orientations
oris = Rotation.random(1000)
Crotated = C*oris  # Compute the Voigt average
Cvoigt = Crotated.Voigt_average()
print(Cvoigt.Young_modulus) # Look at the corresponding Young moduli
print(C.Voigt_average().Young_modulus) # Compare with infinite number of orientations


app.exec() # Only required in PyCharm (see lines 7-10)


import warnings
from Elasticipy.tensors.elasticity import StiffnessTensor as NewStiffnessTensor
from Elasticipy.tensors.elasticity import ComplianceTensor as NewComplianceTensor

warnings.warn(
    "The module 'Elasticipy.FourthOrderTensor' is deprecated and will be removed in a future release. "
    "Please use 'Elasticipy.tensors.elasticity' instead.",
    DeprecationWarning,
    stacklevel=2
)

class StiffnessTensor(NewStiffnessTensor):
    pass

class ComplianceTensor(NewComplianceTensor):
    pass



import warnings
from Elasticipy.tensors.stress_strain import StressTensor as NewStressTensor
from Elasticipy.tensors.stress_strain import StrainTensor as NewStrainTensor

warnings.warn(
    "The module 'Elasticipy.StressStrainTensors' is deprecated and will be removed in a future release. "
    "Please use 'Elasticipy.tensors.stress_strain' instead.",
    DeprecationWarning,
    stacklevel=2
)

class StressTensor(NewStressTensor):
    pass

class StrainTensor(NewStrainTensor):
    pass
